#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.op_return_data: Bitcoin OP_RETURN data class
"""

from ....obj import InitErrors

class OpReturnData(bytes, InitErrors):

	def __new__(cls, proto, data_spec):

		desc = 'OpReturnData initializer'

		assert isinstance(data_spec, str), f'{desc} must be a string'

		if data_spec.startswith('hexdata:'):
			hexdata = data_spec[8:]
			from ....util import is_hex_str
			assert is_hex_str(hexdata), f'{hexdata!r}: {desc} hexdata not in hexadecimal format'
			assert not len(hexdata) % 2, f'{len(hexdata)}: {desc} hexdata of non-even length'
			ret = bytes.fromhex(hexdata)
		elif data_spec.startswith('data:'):
			try:
				ret = data_spec[5:].encode('utf8')
			except:
				raise ValueError(f'{desc} string be UTF-8 encoded')
		else:
			raise ValueError(f'{desc} string must start with ‘data:’ or ‘hexdata:’')

		assert 1 <= len(ret) <= proto.max_op_return_data_len, (
			f'{len(ret)}: {desc} string encoded byte length not in range 1-{proto.max_op_return_data_len}')

		return bytes.__new__(cls, ret)

	def __repr__(self):
		'return an initialization string'
		ret = str(self)
		return ('hexdata:' if self.display_hex else 'data:') + ret

	def __str__(self):
		'return something suitable for display to the user'
		self.display_hex = True
		try:
			ret = self.decode('utf8')
		except:
			return self.hex()
		else:
			import unicodedata
			for ch in ret:
				if ch == '\n' or unicodedata.category(ch)[0] in ('C', 'M'): # see MMGenLabel
					return self.hex()
			self.display_hex = False
			return ret

	def hl(self, *, add_label=False):
		'colorize and optionally label the result of str()'
		from ....color import blue, pink
		ret = str(self)
		if add_label:
			return blue('OP_RETURN data' + (' (hex): ' if self.display_hex else ': ')) + pink(ret)
		else:
			return pink(ret)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.local: local RPC client class for the MMGen Project
"""

import sys, json, asyncio, importlib

from ..util import msg, die, fmt, oneshot_warning

from . import util

class RPCClient:

	is_remote = False
	auth_type = None
	has_auth_cookie = False
	network_proto = 'http'
	proxy = None

	def __init__(self, cfg, host, port, *, test_connection=True):

		self.cfg = cfg
		self.name = type(self).__name__

		# aiohttp workaround, and may speed up RPC performance overall on some systems:
		if sys.platform == 'win32' and host == 'localhost':
			host = '127.0.0.1'

		if not self.cfg.debug_rpc:
			util.dmsg_rpc = util.dmsg_rpc_backend = util.noop

		util.dmsg_rpc(f'=== {self.name}.__init__() debug ===')
		util.dmsg_rpc(f'    cls [{self.name}] host [{host}] port [{port}]\n')

		if test_connection:
			import socket
			try:
				socket.create_connection((host, port), timeout=1).close()
			except:
				die('SocketError', f'Unable to connect to {host}:{port}')

		self.http_hdrs = {'Content-Type': 'application/json'}
		self.host_url = f'{self.network_proto}://{host}:{port}'
		self.host = host
		self.port = port
		self.timeout = self.cfg.http_timeout or 60
		self.auth = None

	def _get_backend(self, backend):
		dfl_backends = {
			'linux': 'httplib',
			'darwin': 'httplib',
			'win32': 'requests'}
		def get_cls(backend_id):
			return getattr(importlib.import_module(f'mmgen.rpc.backends.{backend_id}'), backend_id)
		backend_id = backend or self.cfg.rpc_backend
		return get_cls(dfl_backends[sys.platform] if backend_id == 'auto' else backend_id)(self)

	def set_backend(self, backend=None):
		self.backend = self._get_backend(backend)

	async def set_backend_async(self, backend=None):
		ret = self._get_backend(backend)
		self.backend = (await ret) if type(ret).__name__ == 'coroutine' else ret

	# Call family of methods - direct-to-daemon RPC call:
	# - positional params are passed to the daemon, 'timeout' and 'wallet' kwargs to the backend
	# - 'wallet' kwarg is used only by regtest

	async def call(self, method, *params, timeout=None, wallet=None):
		"""
		default call: call with param list unrolled, exactly as with cli
		"""
		return self.process_http_resp(await self.backend.run(
			payload = {'id': 1, 'jsonrpc': '2.0', 'method': method, 'params': params},
			timeout = timeout,
			host_path = self.make_host_path(wallet)
		))

	async def batch_call(self, method, param_list, *, timeout=None, wallet=None):
		"""
		Make a single call with a list of tuples as first argument
		For RPC calls that return a list of results
		"""
		return self.process_http_resp(await self.backend.run(
			payload = [{
				'id': n,
				'jsonrpc': '2.0',
				'method': method,
				'params': params} for n, params in enumerate(param_list, 1)],
			timeout = timeout,
			host_path = self.make_host_path(wallet)
		), batch=True)

	async def gathered_call(self, method, args_list, *, timeout=None, wallet=None):
		"""
		Perform multiple RPC calls, returning results in a list
		Can be called two ways:
		  1) method = methodname, args_list = [args_tuple1, args_tuple2,...]
		  2) method = None, args_list = [(methodname1, args_tuple1), (methodname2, args_tuple2), ...]
		"""
		cmd_list = args_list if method is None else tuple(zip([method] * len(args_list), args_list))

		cur_pos = 0
		chunk_size = 1024
		ret = []

		while cur_pos < len(cmd_list):
			tasks = [self.backend.run(
						payload = {'id': n, 'jsonrpc': '2.0', 'method': method, 'params': params},
						timeout = timeout,
						host_path = self.make_host_path(wallet)
					) for n, (method, params)  in enumerate(cmd_list[cur_pos:chunk_size+cur_pos], 1)]
			ret.extend(await asyncio.gather(*tasks))
			cur_pos += chunk_size

		return [self.process_http_resp(r) for r in ret]

	# Icall family of methods - indirect RPC call using CallSigs mechanism:
	# - 'timeout' and 'wallet' kwargs are passed to corresponding Call method
	# - remaining kwargs are passed to CallSigs method
	# - CallSigs method returns method and positional params for Call method

	def icall(self, method, **kwargs):
		timeout = kwargs.pop('timeout', None)
		wallet = kwargs.pop('wallet', None)
		return self.call(
			*getattr(self.call_sigs, method)(**kwargs),
			timeout = timeout,
			wallet = wallet)

	def gathered_icall(self, method, args_list, *, timeout=None, wallet=None):
		return self.gathered_call(
			method,
			[getattr(self.call_sigs, method)(*a)[1:] for a in args_list],
			timeout = timeout,
			wallet = wallet)

	def process_http_resp(self, run_ret, *, batch=False, json_rpc=True):

		def float_parser(n):
			return n

		text, status = run_ret

		if status == 200:
			util.dmsg_rpc('    RPC RESPONSE data ==>\n{}\n', text, is_json=True)
			m = None
			if batch:
				return [r['result'] for r in json.loads(text, parse_float=float_parser)]
			else:
				try:
					if json_rpc:
						ret = json.loads(text, parse_float=float_parser)['result']
						if isinstance(ret, list) and ret and type(ret[0]) == dict and 'success' in ret[0]:
							for res in ret:
								if not res['success']:
									m = str(res['error'])
									assert False
						return ret
					else:
						return json.loads(text, parse_float=float_parser)
				except:
					if not m:
						t = json.loads(text)
						try:
							m = t['error']['message']
						except:
							try:
								m = t['error']
							except:
								m = t
					die('RPCFailure', m)
		else:
			import http
			m, s = ('', http.HTTPStatus(status))
			if text:
				try:
					m = json.loads(text)['error']['message']
				except:
					try:
						m = text.decode()
					except:
						m = text
			die('RPCFailure', f'{s.value} {s.name}: {m}')

	async def stop_daemon(self, *, quiet=False, silent=False):
		if self.daemon.state == 'ready':
			if not (quiet or silent):
				msg(f'Stopping {self.daemon.desc} on port {self.daemon.bind_port}')
			ret = await self.do_stop_daemon(silent=silent)
			if self.daemon.wait:
				self.daemon.wait_for_state('stopped')
			return ret
		else:
			if not (quiet or silent):
				msg(f'{self.daemon.desc} on port {self.daemon.bind_port} not running')
			return True

	def start_daemon(self, *, silent=False):
		return self.daemon.start(silent=silent)

	async def restart_daemon(self, *, quiet=False, silent=False):
		await self.stop_daemon(quiet=quiet, silent=silent)
		return self.daemon.start(silent=silent)

	def handle_unsupported_daemon_version(self, name, warn_only):

		class daemon_version_warning(oneshot_warning):
			color = 'yellow'
			message = 'ignoring unsupported {} daemon version at user request'

		if warn_only:
			daemon_version_warning(div=name, fmt_args=[self.daemon.coind_name])
		else:
			name = self.daemon.coind_name
			die(2, '\n'+fmt(f"""
				The running {name} daemon has version {self.daemon_version_str}.
				This version of MMGen is tested only on {name} v{self.daemon.coind_version_str} and below.

				To avoid this error, downgrade your daemon to a supported version.

				Alternatively, you may invoke the command with the --ignore-daemon-version
				option, in which case you proceed at your own risk.
				""", indent='    '))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.chainsplit: Forking scenario tests for the cmdtest.py test suite
This module is unmaintained and currently non-functional
"""

from mmgen.util import die

from .include.common import get_file_with_ext, rt_pw
from .regtest import CmdTestRegtest

class CmdTestChainsplit(CmdTestRegtest):
	'forking scenario tests for the cmdtest.py test suite'
	cmd_group = (
		('split_setup',                  'regtest forking scenario setup'),
		('walletgen_bob',                'generating Bob’s wallet'),
		('addrgen_bob',                  'generating Bob’s addresses'),
		('addrimport_bob',               'importing Bob’s addresses'),
		('fund_bob',                     'funding Bob’s wallet'),
		('split_fork',                   'regtest split fork'),
		('split_start_btc',              'start regtest daemon (BTC)'),
		('split_start_b2x',              'start regtest daemon (B2X)'),
		('split_gen_btc',                'mining a block (BTC)'),
		('split_gen_b2x',                'mining 100 blocks (B2X)'),
		('split_do_split',               'creating coin splitting transactions'),
		('split_sign_b2x',               'signing B2X split transaction'),
		('split_sign_btc',               'signing BTC split transaction'),
		('split_send_b2x',               'sending B2X split transaction'),
		('split_send_btc',               'sending BTC split transaction'),
		('split_gen_btc',                'mining a block (BTC)'),
		('split_gen_b2x2',               'mining a block (B2X)'),
		('split_txdo_timelock_bad_btc',  'sending transaction with bad locktime (BTC)'),
		('split_txdo_timelock_good_btc', 'sending transaction with good locktime (BTC)'),
		('split_txdo_timelock_bad_b2x',  'sending transaction with bad locktime (B2X)'),
		('split_txdo_timelock_good_b2x', 'sending transaction with good locktime (B2X)'),
	)

	def split_setup(self):
		if self.proto.coin != 'BTC':
			die(1, 'Test valid only for coin BTC')
		self.coin = 'BTC'
		return self.setup()

	def split_fork(self):
		self.coin = 'B2X'
		t = self.spawn('mmgen-regtest', ['fork', 'btc'])
		t.expect('Creating fork from coin')
		t.expect('successfully created')
		t.ok()

	def split_start(self, coin):
		self.coin = coin
		t = self.spawn('mmgen-regtest', ['bob'])
		t.expect('Starting')
		t.expect('done')
		t.ok()

	def split_start_btc(self):
		self.regtest_start(coin='BTC')

	def split_start_b2x(self):
		self.regtest_start(coin='B2X')

	def split_gen_btc(self):
		self.regtest_generate(coin='BTC')

	def split_gen_b2x(self):
		self.regtest_generate(coin='B2X', num_blocks=100)

	def split_gen_b2x2(self):
		self.regtest_generate(coin='B2X')

	def split_do_split(self):
		self.coin = 'B2X'
		sid = self.regtest_user_sid('bob')
		t = self.spawn('mmgen-split', [
			'--bob',
			'--outdir='+self.tmpdir,
			'--tx-fees=0.0001,0.0003',
			sid+':S:1', sid+':S:2'])
		t.expect(r'\[q\]uit menu, .*?:.', 'q', regex=True)
		t.expect('outputs to spend: ', '1\n')

		for _ in ('timelocked', 'split'):
			for _ in ('fee', 'change'):
				t.expect('OK? (Y/n): ', 'y')
			t.do_comment(False)
			t.view_tx('t')

		t.written_to_file('Long chain (timelocked) transaction')
		t.written_to_file('Short chain transaction')
		t.ok()

	def split_sign(self, coin, ext):
		wf = get_file_with_ext(self.regtest_user_dir('bob', coin=coin.lower()), 'mmdat')
		txfile = self.get_file_with_ext(ext, no_dot=True)
		self.coin = coin
		self.txsign(txfile, wf, extra_opts=['--bob'])

	def split_sign_b2x(self):
		return self.regtest_sign(coin='B2X', ext='533].rawtx')

	def split_sign_btc(self):
		return self.regtest_sign(coin='BTC', ext='9997].rawtx')

	def split_send(self, coin, ext):
		self.coin = coin
		txfile = self.get_file_with_ext(ext, no_dot=True)
		self.txsend(txfile, bogus_send=False, extra_opts=['--bob'])

	def split_send_b2x(self):
		return self.regtest_send(coin='B2X', ext='533].sigtx')

	def split_send_btc(self):
		return self.regtest_send(coin='BTC', ext='9997].sigtx')

	def split_txdo_timelock(self, coin, locktime, bad_locktime):
		self.coin = coin
		sid = self.regtest_user_sid('bob')
		self.regtest_user_txdo(
			'bob',
			'0.0001',
			[sid+':S:5'],
			'1',
			pw           = rt_pw,
			extra_args   = ['--locktime='+str(locktime)],
			bad_locktime = bad_locktime)

	def split_txdo_timelock_bad_btc(self):
		self.regtest_txdo_timelock('BTC', locktime=8888, bad_locktime=True)
	def split_txdo_timelock_good_btc(self):
		self.regtest_txdo_timelock('BTC', locktime=1321009871, bad_locktime=False)
	def split_txdo_timelock_bad_b2x(self):
		self.regtest_txdo_timelock('B2X', locktime=8888, bad_locktime=True)
	def split_txdo_timelock_good_b2x(self):
		self.regtest_txdo_timelock('B2X', locktime=1321009871, bad_locktime=False)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xmr.rpc: Monero base protocol RPC client class
"""

import re

from ...rpc.local import RPCClient
from ...rpc.util import IPPort, auth_data

class MoneroRPCClient(RPCClient):

	auth_type = None
	network_proto = 'https'
	verify_server = False

	def __init__(
			self,
			cfg,
			proto,
			*,
			host,
			port,
			user,
			passwd,
			test_connection       = True,
			proxy                 = None,
			daemon                = None,
			ignore_daemon_version = False):

		self.proto = proto

		if proxy is not None:
			self.proxy = IPPort(proxy)
			test_connection = False
			if host.endswith('.onion'):
				self.network_proto = 'http'

		super().__init__(cfg, host, port, test_connection=test_connection)

		if self.auth_type:
			self.auth = auth_data(user, passwd)

		if True:
			self.set_backend('requests')

		else: # insecure, for debugging only
			self.set_backend('curl')
			self.backend.exec_opts.remove('--silent')
			self.backend.exec_opts.append('--verbose')

		self.daemon = daemon

		if test_connection:
			# https://github.com/monero-project/monero src/rpc/rpc_version_str.cpp
			ver_str = self.call_raw('getinfo')['version']
			if ver_str:
				self.daemon_version_str = ver_str
				self.daemon_version = sum(
					int(m) * (1000 ** n) for n, m in
						enumerate(reversed(re.match(r'(\d+)\.(\d+)\.(\d+)\.(\d+)', ver_str).groups()))
				)
				if self.daemon and self.daemon_version > self.daemon.coind_version:
					self.handle_unsupported_daemon_version(
						proto.name,
						ignore_daemon_version or proto.ignore_daemon_version or self.cfg.ignore_daemon_version)
			else: # restricted (public) node:
				self.daemon_version_str = None
				self.daemon_version = None

	def call(self, method, *params, **kwargs):
		assert not params, f'{self.name}.call() accepts keyword arguments only'
		return self.process_http_resp(self.backend.run_noasync(
			payload = {'id': 0, 'jsonrpc': '2.0', 'method': method, 'params': kwargs},
			timeout = 3600, # allow enough time to sync ≈1,000,000 blocks
			host_path = '/json_rpc'
		))

	def call_raw(self, method, *params, **kwargs):
		assert not params, f'{self.name}.call() accepts keyword arguments only'
		return self.process_http_resp(self.backend.run_noasync(
			payload = kwargs,
			timeout = self.timeout,
			host_path = f'/{method}'
		), json_rpc=False)

	async def do_stop_daemon(self, *, silent=False):
		return self.call_raw('stop_daemon') # unreliable on macOS (daemon stops, but closes connection)

	rpcmethods = ('get_info',)
	rpcmethods_raw = ('get_height', 'send_raw_transaction', 'stop_daemon')

class MoneroWalletRPCClient(MoneroRPCClient):

	auth_type = 'digest'

	def __init__(self, cfg, daemon, *, test_connection=True):

		RPCClient.__init__(
			self            = self,
			cfg             = cfg,
			host            = 'localhost',
			port            = daemon.rpc_port,
			test_connection = test_connection)

		self.daemon = daemon
		self.auth = auth_data(daemon.user, daemon.passwd)
		self.set_backend('requests')

	rpcmethods = (
		'get_version',
		'get_height',    # sync height of the open wallet
		'get_balance',   # account_index=0, address_indices=[]
		'create_wallet', # filename, password, language="English"
		'open_wallet',   # filename, password
		'close_wallet',
		# filename,password,seed (restore_height,language,seed_offset,autosave_current)
		'restore_deterministic_wallet',
		'refresh',       # start_height
	)

	def call_raw(self, *args, **kwargs):
		raise NotImplementedError('call_raw() not implemented for class MoneroWalletRPCClient')

	async def do_stop_daemon(self, *, silent=False):
		"""
		NB: the 'stop_wallet' RPC call closes the open wallet before shutting down the daemon,
		returning an error if no wallet is open
		"""
		try:
			return self.call('stop_wallet')
		except Exception as e:
			from ...util import msg, msg_r, ymsg
			from ...color import yellow
			msg(f'{type(e).__name__}: {e}')
			msg_r(yellow('Unable to shut down wallet daemon gracefully, so killing process instead...'))
			ret = self.daemon.stop(silent=True)
			ymsg('done')
			return ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.swap: asset swap tests for the cmdtest.py test suite
"""

from pathlib import Path

from mmgen.cfg import Config
from mmgen.protocol import init_proto
from mmgen.wallet.mmgen import wallet as MMGenWallet

from ..include.common import imsg, make_burn_addr, gr_uc

from .include.runner import CmdTestRunner
from .include.common import dfl_bip39_file, dfl_words_file
from .httpd.thornode.swap import ThornodeSwapServer

from .autosign import CmdTestAutosign, CmdTestAutosignThreaded
from .regtest import CmdTestRegtest, rt_data, dfl_wcls, rt_pw, strip_ansi_escapes

sample1 = gr_uc[:24]
sample2 = '00010203040506'

def create_cross_methods(cross_coin, cross_group, cmd_group_in, cmd_subgroups):

	method_template = """
def {name}(self):
	self.spawn(log_only=True)
	return {group}.run_test("{method_name}", sub=True)
"""

	tests = [c[0] for v in tuple(cmd_subgroups.values()) + (cmd_group_in,)
		for c in v if isinstance(c, tuple) and c[0].startswith(f'{cross_coin}_')]

	return ''.join(method_template.format(
		name = k,
		method_name = k.removeprefix(f'{cross_coin}_'),
		group = cross_group)
			for k in tests)

class CmdTestSwapMethods:

	@property
	def bob_opt(self):
		return ['--bob'] if self.proto.base_proto == 'Bitcoin' else ['--regtest=1']

	@property
	def fee_desc(self):
		return 'fee or gas price' if self.proto.is_evm else 'fee'

	def walletconv_bob(self):
		t = self.spawn(
			'mmgen-walletconv',
			['--bob', '--quiet', '-r0', f'-d{self.cfg.data_dir}/regtest/bob', dfl_words_file],
			no_passthru_opts = ['coin', 'eth_daemon_id'])
		t.hash_preset(MMGenWallet.desc, '1')
		t.passphrase_new('new '+MMGenWallet.desc, rt_pw)
		t.label()
		return t

	def _addrgen_bob(self, proto_idx, mmtypes, subseed_idx=None):
		return self.addrgen('bob', subseed_idx=subseed_idx, mmtypes=mmtypes, proto=self.protos[proto_idx])

	def _addrimport_bob(self, proto_idx):
		return self.addrimport('bob', mmtypes=['S', 'B'], proto=self.protos[proto_idx])

	def _fund_bob(self, proto_idx, addrtype_code, amt):
		return self.fund_wallet('bob', amt, mmtype=addrtype_code, proto=self.protos[proto_idx])

	def _bob_bal(self, proto_idx, bal, skip_check=False):
		return self.user_bal('bob', bal, proto=self.protos[proto_idx], skip_check=skip_check)

	def _data_tx_create(self, src, dest, chg, pfx, sample):
		t = self.spawn(
			'mmgen-txcreate',
			['-d', self.tmpdir, '-B', '--bob', f'{self.sid}:{dest},1', f'{self.sid}:{chg}', f'{pfx}:{sample}'])
		return self.txcreate_ui_common(t, menu=[], inputs='1', interactive_fee='3s')

	def _data_tx_sign(self):
		fn = self.get_file_with_ext('rawtx')
		t = self.spawn('mmgen-txsign', ['-d', self.tmpdir, '--bob', fn])
		t.view_tx('v')
		t.passphrase(dfl_wcls.desc, rt_pw)
		t.do_comment(None)
		t.expect('(Y/n): ', 'y')
		t.written_to_file('Signed transaction')
		return t

	def _data_tx_send(self):
		fn = self.get_file_with_ext('sigtx')
		t = self.spawn('mmgen-txsend', ['-q', '-d', self.tmpdir, '--bob', fn])
		t.expect('view: ', 'n')
		t.expect('(y/N): ', '\n')
		t.expect('to confirm: ', 'YES\n')
		t.written_to_file('Sent transaction')
		return t

	def _data_tx_do(self, src, dest, chg, pfx, sample, view):
		t = self.user_txdo(
			user         = 'bob',
			fee          = '30s',
			outputs_cl   = [f'{self.sid}:{dest},1', f'{self.sid}:{chg}', f'{pfx}:{sample}'],
			outputs_list = src,
			add_comment  = 'Transaction with OP_RETURN data',
			return_early = True)

		t.view_tx(view)
		if view == 'v':
			t.expect(sample)
			t.expect('amount:')
		t.passphrase(dfl_wcls.desc, rt_pw)
		t.written_to_file('Signed transaction')
		self._do_confirm_send(t)
		t.expect('Transaction sent')
		return t

	def _data_tx_chk(self, sample):
		mp = self._get_mempool(do_msg=True)
		assert len(mp) == 1
		self.write_to_tmpfile('data_tx1_id', mp[0]+'\n')
		tx_hex = self._do_cli(['getrawtransaction', mp[0]])
		tx = self._do_cli(['decoderawtransaction', tx_hex], decode_json=True)
		v0 = tx['vout'][0]
		assert v0['scriptPubKey']['hex'] == f'6a{(len(sample) // 2):02x}{sample}'
		assert v0['scriptPubKey']['type'] == 'nulldata'
		assert v0['value'] == "0.00000000"
		return 'ok'

	def _swaptxcreate_ui_common(
			self,
			t,
			*,
			inputs          = '1',
			interactive_fee = None,
			file_desc       = 'Unsigned transaction',
			tweaks          = [],
			reload_quote    = False,
			sign_and_send   = False,
			need_passphrase = True,
			expect         = None):
		t.expect(self.menu_prompt, 'q')
		t.expect(self.input_sels_prompt, f'{inputs}\n')
		if reload_quote:
			t.expect('to continue: ', 'r')  # reload swap quote
		t.expect('to continue: ', '\n')     # exit swap quote view
		if self.proto.has_usr_fee:
			t.expect('(Y/n): ', 'y')            # fee OK?
			t.expect('(Y/n): ', 'y')            # change OK?
		t.expect('(y/N): ', 'n')            # add comment?
		t.expect('view: ', 'y')             # view TX
		if expect:
			t.expect(expect)
		t.expect('to continue: ', '\n')
		if sign_and_send:
			if need_passphrase:
				t.passphrase(dfl_wcls.desc, rt_pw)
			t.expect('to confirm: ', 'YES\n')
		else:
			t.expect('(y/N): ', 'y')            # save?
		t.written_to_file(file_desc)
		return t

	def _swaptxcreate(self, args, *, action='txcreate', add_opts=[], exit_val=None):
		self.get_file_with_ext('rawtx', delete_all=True)
		return self.spawn(
			f'mmgen-swap{action}',
			['-q', '-d', self.tmpdir, '-B', '--bob']
			+ add_opts
			+ args,
			exit_val = exit_val,
			no_passthru_opts = ['coin'])

	def _swaptxcreate_bad(self, args, *, exit_val=1, expect1=None, expect2=None):
		t = self._swaptxcreate(args, exit_val=exit_val)
		if expect1:
			t.expect(expect1)
		if expect2:
			t.expect(expect2)
		return t

	def _swaptxsend1(self, *, add_opts=[], spawn_only=False):
		return self._swaptxsend(
			add_opts = add_opts + [
			# test overriding host:port with coin-specific options:
			'--rpc-host=unreachable', # unreachable host
			'--ltc-rpc-host=localhost',
			'--rpc-port=46381',       # bad port
			'--ltc-rpc-port=20680',
			],
			spawn_only = spawn_only)

	def _swaptxsend(self, *, add_opts=[], spawn_only=False, status=False, dump_hex=False):
		fn = self.get_file_with_ext('sigtx')
		t = self.spawn(
			'mmgen-txsend',
			add_opts + ['-q', '-d', self.tmpdir, '--bob', fn],
			no_passthru_opts = ['coin'])
		if spawn_only:
			return t
		t.view_tx('v')
		if status:
			return t
		t.expect('(y/N): ', 'n')
		if not dump_hex:
			t.expect('to confirm: ', 'YES\n')
		return t

	def _swaptxsign(self, *, add_opts=[], expect=None):
		self.get_file_with_ext('sigtx', delete_all=True)
		fn = self.get_file_with_ext('rawtx')
		t = self.spawn(
			'mmgen-txsign',
			add_opts + ['-d', self.tmpdir, '--bob', fn],
			no_passthru_opts = ['coin'])
		t.view_tx('t')
		if expect:
			t.expect(expect)
		t.passphrase(dfl_wcls.desc, rt_pw)
		t.do_comment(None)
		t.expect('(Y/n): ', 'y')
		t.written_to_file('Signed transaction')
		return t

	def _swaptxbump(self, fee, *, add_opts=[], output_args=[], exit_val=None):
		self.get_file_with_ext('rawtx', delete_all=True)
		fn = self.get_file_with_ext('sigtx')
		t = self.spawn(
			'mmgen-txbump',
			['-q', '-d', self.tmpdir] + self.bob_opt + add_opts + output_args + [fn],
			exit_val = exit_val)
		return self._swaptxbump_ui_common(t, interactive_fee=fee, new_outputs=bool(output_args))

	def _swaptxbump_ui_common_new_outputs(
			self,
			t,
			*,
			tweaks          = [],
			inputs          = None,
			interactive_fee = None,
			file_desc       = None):
		return self._swaptxbump_ui_common(
				t,
				inputs          = inputs,
				interactive_fee = interactive_fee,
				file_desc       = file_desc,
				new_outputs     = True)

	def _swaptxbump_ui_common(
			self,
			t,
			*,
			inputs          = None,
			interactive_fee = None,
			file_desc       = None,
			new_outputs     = False):
		if new_outputs:
			if not self.proto.is_vm:
				t.expect(f'{self.fee_desc}: ', interactive_fee + '\n')
			t.expect('(Y/n): ', 'y')        # fee ok?
			t.expect('(Y/n): ', 'y')        # change ok?
		else:
			if not self.proto.is_vm:
				t.expect('ENTER for the change output): ', '\n')
				t.expect('(Y/n): ', 'y')    # confirm deduct from chg output
			t.expect('to continue: ', '\n') # exit swap quote
			t.expect(f'{self.fee_desc}: ', interactive_fee + '\n')
			t.expect('(Y/n): ', 'y')        # fee ok?
		t.expect('(y/N): ', 'n')            # comment?
		t.expect('(y/N): ', 'y')            # save?
		return t

	def _generate_for_proto(self, proto_idx):
		return self.generate(num_blocks=1, add_opts=[f'--coin={self.protos[proto_idx].coin}'])

	def _swaptxsign_bad(self, expect, *, add_opts=[], exit_val=1):
		self.get_file_with_ext('sigtx', delete_all=True)
		fn = self.get_file_with_ext('rawtx')
		t = self.spawn('mmgen-txsign', add_opts + ['-d', self.tmpdir, '--bob', fn], exit_val=exit_val)
		t.expect('view: ', '\n')
		t.expect(expect)
		return t

	def _listaddresses(self, proto_idx):
		return self.user_bal('bob', None, proto=self.protos[proto_idx], skip_check=True)

	def _mempool(self, proto_idx):
		self.spawn(msg_only=True)
		data = self._do_cli(['getrawmempool'], add_opts=[f'--coin={self.protos[proto_idx].coin}'])
		assert data
		return 'ok'

	def _thornode_server_stop(self, attrname='swap_server', name='thornode swap server'):
		self.spawn(msg_only=True)
		if self.cfg.no_daemon_stop:
			imsg(f'(leaving {name} running by user request)')
		else:
			getattr(self, attrname).stop()
		return 'ok'

	def create_cross_runner(self, trunner, *, add_cfg={}):
		cfg = Config({
			'_clone': trunner.cfg,
			'coin': self.cross_coin,
			'resume': None,
			'resuming': None,
			'resume_after': None,
			'exit_after': None,
			'log': None} | add_cfg)
		t = trunner
		ret = CmdTestRunner(cfg, t.repo_root, t.data_dir, t.trash_dir, t.trash_dir2)
		ret.init_group(self.cross_group)
		return ret

class CmdTestSwap(CmdTestSwapMethods, CmdTestRegtest, CmdTestAutosignThreaded):
	'swap operations for BTC, BCH and LTC'

	bdb_wallet = True
	networks = ('btc',)
	tmpdir_nums = [37]
	passthru_opts = ('rpc_backend',)
	coins = ('btc',)
	need_daemon = True

	cmd_group_in = (
		('list_assets',           'listing swap assets'),
		('subgroup.init_data',    []),
		('subgroup.data',         ['init_data']),
		('subgroup.init_swap',    []),
		('subgroup.create',       ['init_swap']),
		('subgroup.create_bad',   ['init_swap']),
		('subgroup.signsend',     ['init_swap']),
		('subgroup.signsend_bad', ['init_swap']),
		('subgroup.autosign',     ['init_data', 'signsend']),
		('swap_server_stop',      'stopping the Thornode server'),
		('stop',                  'stopping regtest daemons'),
	)
	cmd_subgroups = {
		'init_data': (
			'Initialize regtest setup for OP_RETURN data operations',
			('setup',            'regtest (Bob and Alice) mode setup'),
			('walletcreate_bob', 'wallet creation (Bob)'),
			('addrgen_bob',      'address generation (Bob)'),
			('addrimport_bob',   'importing Bob’s addresses'),
			('fund_bob1',        'funding Bob’s wallet (bech32)'),
			('fund_bob2',        'funding Bob’s wallet (native Segwit)'),
			('bob_bal',          'displaying Bob’s balance'),
		),
		'data': (
			'OP_RETURN data operations',
			('data_tx1_create',  'Creating a transaction with OP_RETURN data (hex-encoded UTF-8)'),
			('data_tx1_sign',    'Signing the transaction'),
			('data_tx1_send',    'Sending the transaction'),
			('data_tx1_chk',     'Checking the sent transaction'),
			('generate3',        'Generate 3 blocks'),
			('data_tx2_do',      'Creating and sending a transaction with OP_RETURN data (binary)'),
			('data_tx2_chk',     'Checking the sent transaction'),
			('generate3',        'Generate 3 blocks'),
			('bob_listaddrs',    'Display Bob’s addresses'),
		),
		'init_swap': (
			'Initialize regtest setup for swap operations',
			('setup_send_coin',               'setting up the sending coin regtest blockchain'),
			('walletcreate_bob',              'wallet creation (Bob)'),
			('addrgen_bob_send',              'address generation (Bob, sending coin)'),
			('addrimport_bob_send',           'importing Bob’s addresses (sending coin)'),
			('fund_bob_send',                 'funding Bob’s wallet (bech32)'),
			('bob_bal_send',                  'displaying Bob’s send balance'),

			('setup_recv_coin',               'setting up the receiving coin regtest blockchain'),
			('addrgen_bob_recv',              'address generation (Bob, receiving coin)'),
			('addrimport_bob_recv',           'importing Bob’s addresses (receiving coin)'),
			('fund_bob_recv1',                'funding Bob’s wallet (bech32)'),
			('fund_bob_recv2',                'funding Bob’s wallet (native Segwit)'),
			('addrgen_bob_recv_subwallet',    'address generation (Bob, receiving coin)'),
			('addrimport_bob_recv_subwallet', 'importing Bob’s addresses (receiving coin)'),
			('fund_bob_recv_subwallet',       'funding Bob’s subwwallet (native Segwit)'),
			('bob_bal_recv',                  'displaying Bob’s receive balance'),
		),
		'create': (
			'Swap TX create operations (BCH => LTC)',
			('swaptxcreate1',  'creating a swap transaction (full args)'),
			('swaptxcreate2',  'creating a swap transaction (coin args only)'),
			('swaptxcreate3',  'creating a swap transaction (no chg arg)'),
			('swaptxcreate4',  'creating a swap transaction (chg and dest by addrtype)'),
			('swaptxcreate5',  'creating a swap transaction (chg and dest by addrlist ID)'),
			('swaptxcreate6',  'creating a swap transaction (dest is non-wallet addr)'),
			('swaptxcreate7',  'creating a swap transaction (coin-amt-coin)'),
		),
		'create_bad': (
			'Swap TX create operations: error handling',
			('swaptxcreate_bad1',  'creating a swap transaction (bad, used destination address)'),
			('swaptxcreate_bad2',  'creating a swap transaction (bad, used change address)'),
			('swaptxcreate_bad3',  'creating a swap transaction (bad, unsupported send coin)'),
			('swaptxcreate_bad4',  'creating a swap transaction (bad, unsupported recv coin)'),
			('swaptxcreate_bad5',  'creating a swap transaction (bad, malformed cmdline)'),
			('swaptxcreate_bad6',  'creating a swap transaction (bad, malformed cmdline)'),
			('swaptxcreate_bad7',  'creating a swap transaction (bad, bad user input, user exit)'),
			('swaptxcreate_bad8',  'creating a swap transaction (bad, non-MMGen change address)'),
			('swaptxcreate_bad9',  'creating a swap transaction (bad, invalid addrtype)'),
		),
		'signsend': (
			'Swap TX create, sign and send operations (LTC => BCH)',
			('swaptxsign1_create', 'creating a swap transaction (full args)'),
			('swaptxsign1',        'signing the transaction'),
			('swaptxsend1',        'sending the transaction'),
			('swaptxsend1_status', 'getting status of sent transaction'),
			('generate1',          'generating a block'),
			('swaptxsign2_create', 'creating a swap transaction (non-wallet swap address)'),
			('swaptxsign2',        'signing the transaction'),
			('swaptxsend2',        'sending the transaction'),
			('mempool1',           'viewing the mempool'),
			('swaptxbump1',        'bumping the transaction'),
			('swaptxsign3',        'signing the transaction'),
			('swaptxsend3',        'sending the transaction'),
			('mempool1',           'viewing the mempool'),
			('swaptxbump2',        'bumping the transaction again'),
			('swaptxsign4',        'signing the transaction'),
			('swaptxsend4',        'sending the transaction'),
			('mempool1',           'viewing the mempool'),
			('generate1',          'generating a block'),
			('swap_bal1',          'checking the balance'),
			('swaptxsign1_do',     'creating, signing and sending a swap transaction'),
			('generate1',          'generating a block'),
			('swap_bal2',          'checking the balance'),
		),
		'signsend_bad': (
			'Swap TX create, sign and send operations: error handling',
			('swaptxsign_bad1_create', 'creating a swap transaction (non-wallet swap address)'),
			('swaptxsign_bad1',        'signing the transaction (non-wallet swap address)'),
			('swaptxsign_bad2_create', 'creating a swap transaction'),
			('swaptxsign_bad2',        'signing the transaction'),
			('swaptxsend_bad2',        'sending the transaction (swap quote expired)'),
		),
		'autosign': (
			'Swap TX operations with autosigning (BTC => LTC)',
			('run_setup_bip39',        'setting up offline autosigning'),
			('swap_wait_loop_start',   'starting autosign wait loop'),
			('autosign_swaptxcreate1', 'creating a swap transaction'),
			('autosign_swaptxsend1',   'sending the transaction'),
			('autosign_swaptxbump1',   'bumping the transaction'),
			('autosign_swaptxsend2',   'sending the transaction'),
			('generate0',              'generating a block'),
			('swap_bal3',              'checking the balance'),
			('wait_loop_kill',         'stopping autosign wait loop'),
		),
	}

	def __init__(self, cfg, trunner, cfgs, spawn):

		CmdTestAutosignThreaded.__init__(self, cfg, trunner, cfgs, spawn)
		CmdTestRegtest.__init__(self, cfg, trunner, cfgs, spawn)

		if trunner is None:
			return

		globals_dict = globals()
		for k in rt_data:
			globals_dict[k] = rt_data[k]['btc']

		self.protos = [init_proto(cfg, k, network='regtest', need_amt=True) for k in ('btc', 'ltc', 'bch')]

		self.swap_server = ThornodeSwapServer()
		self.swap_server.start()

		self.opts.append('--bob')

	@property
	def sid(self):
		return self._user_sid('bob')

	def list_assets(self):
		t = self.spawn('mmgen-swaptxcreate', ['--list-assets'])
		t.expect('AVAILABLE')
		t.expect('ETH.MM1')
		t.expect('Blacklisted')
		t.expect('ETH.JUNK')
		return t

	def walletcreate_bob(self):
		dest = Path(self.tr.data_dir, 'regtest', 'bob')
		dest.mkdir(exist_ok=True)
		t = self.spawn('mmgen-walletconv', [
			'--quiet',
			'--usr-randchars=0',
			'--hash-preset=1',
			'--label=SwapWalletLabel',
			f'--outdir={str(dest)}',
			dfl_bip39_file])
		t.expect('wallet: ', rt_pw + '\n')
		t.expect('phrase: ', rt_pw + '\n')
		t.written_to_file('wallet')
		return t

	def addrgen_bob(self):
		return self._addrgen_bob(0, ['S', 'B'])

	def addrimport_bob(self):
		return self._addrimport_bob(0)

	def fund_bob1(self):
		return self._fund_bob(0, 'B', '500')

	def fund_bob2(self):
		return self._fund_bob(0, 'S', '500')

	def bob_bal(self):
		return self._bob_bal(0, '1000')

	def data_tx1_create(self):
		return self._data_tx_create('1', 'B:2', 'B:3', 'data', sample1)

	def data_tx1_sign(self):
		return self._data_tx_sign()

	def data_tx1_send(self):
		return self._data_tx_send()

	def data_tx1_chk(self):
		return self._data_tx_chk(sample1.encode().hex())

	def data_tx2_do(self):
		return self._data_tx_do('2', 'B:4', 'B:5', 'hexdata', sample2, 'v')

	def data_tx2_chk(self):
		return self._data_tx_chk(sample2)

	def generate3(self):
		return self.generate(3)

	def bob_listaddrs(self):
		t = self.spawn('mmgen-tool', ['--bob', 'listaddresses'])
		return t

	def setup_send_coin(self):
		self.user_sids = {}
		return self._setup(proto=self.protos[2], remove_datadir=True)

	def addrgen_bob_send(self):
		return self._addrgen_bob(2, ['C'])

	def addrimport_bob_send(self):
		return self.addrimport('bob', mmtypes=['C'], proto=self.protos[2])

	def fund_bob_send(self):
		return self._fund_bob(2, 'C', '5')

	def bob_bal_send(self):
		return self._bob_bal(2, '5')

	def setup_recv_coin(self):
		return self._setup(proto=self.protos[1], remove_datadir=False)

	def addrgen_bob_recv(self):
		return self._addrgen_bob(1, ['S', 'B'])

	def addrimport_bob_recv(self):
		return self._addrimport_bob(1)

	def fund_bob_recv1(self):
		return self._fund_bob(1, 'S', '5')

	def fund_bob_recv2(self):
		return self._fund_bob(1, 'B', '5')

	def addrgen_bob_recv_subwallet(self):
		return self._addrgen_bob(1, ['C', 'B'], subseed_idx='29L')

	def addrimport_bob_recv_subwallet(self):
		return self._subwallet_addrimport('bob', '29L', ['C', 'B'], proto=self.protos[1])

	def fund_bob_recv_subwallet(self, proto_idx=1, amt='5'):
		coin_arg = f'--coin={self.protos[proto_idx].coin}'
		t = self.spawn('mmgen-tool', ['--bob', coin_arg, 'listaddresses'])
		addr = [s for s in strip_ansi_escapes(t.read()).splitlines() if 'C:1 No' in s][0].split()[3]
		t = self.spawn(
			'mmgen-regtest',
			[coin_arg, 'send', addr, str(amt)],
			no_passthru_opts = ['coin'],
			no_msg = True)
		return t

	def bob_bal_recv(self):
		return self._bob_bal(1, '15')

	def swaptxcreate1(self, idx=3):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(
				['BCH', '1.234', f'{self.sid}:C:{idx}', 'LTC', f'{self.sid}:B:3'],
				add_opts = ['--trade-limit=0%', '--stream-interval=1']),
			expect = ':3541e5/1/0')

	def swaptxcreate2(self):
		t = self._swaptxcreate(
			['BCH', 'LTC'],
			add_opts = ['--no-quiet', '--stream-interval=10', '--trade-limit=3.337%'])
		t.expect('Enter a number> ', '1')
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t, reload_quote=True, expect=':1386e6/10/0')

	def swaptxcreate3(self):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(
				['BCH', 'LTC', f'{self.sid}:B:3'],
				add_opts = ['--trade-limit=10.1%']),
			expect = ':1289e6/3/0')

	def swaptxcreate4(self):
		t = self._swaptxcreate(
			['BCH', '1.234', 'C', 'LTC', 'B'],
			add_opts = ['--trade-limit=-1.123%'])
		t.expect('OK? (Y/n): ', 'y')
		t.expect('Enter a number> ', '1')
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t, expect=':358e6/3/0')

	def swaptxcreate5(self):
		t = self._swaptxcreate(
			['BCH', '1.234', f'{self.sid}:C', 'LTC', f'{self.sid}:B'],
			add_opts = ['--trade-limit=3.6'])
		t.expect('OK? (Y/n): ', 'y')
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t, expect=':36e7/3/0')

	def swaptxcreate6(self):
		addr = make_burn_addr(self.protos[1], mmtype='bech32')
		t = self._swaptxcreate(
			['BCH', '1.234', f'{self.sid}:C', 'LTC', addr],
			add_opts = ['--trade-limit=2.7%'])
		t.expect('OK? (Y/n): ', 'y')
		t.expect('to confirm: ', 'YES\n')
		return self._swaptxcreate_ui_common(t, expect=':3445e5/3/0')

	def swaptxcreate7(self):
		t = self._swaptxcreate(['BCH', '0.56789', 'LTC'])
		t.expect('OK? (Y/n): ', 'y')
		t.expect('Enter a number> ', '1')
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t, expect=':0/3/0')

	def swaptxcreate_bad1(self):
		t = self._swaptxcreate_bad(
			['BCH', '1.234', f'{self.sid}:C:3', 'LTC', f'{self.sid}:S:1'],
			expect1 = 'Requested destination address',
			expect2 = 'Address reuse harms your privacy')
		t.expect('(y/N): ', 'n')
		return t

	def swaptxcreate_bad2(self):
		t = self._swaptxcreate_bad(
			['BCH', '1.234', f'{self.sid}:C:1', 'LTC', f'{self.sid}:S:2'],
			expect1 = 'Requested change address',
			expect2 = 'Address reuse harms your privacy')
		t.expect('(y/N): ', 'n')
		return t

	def swaptxcreate_bad3(self):
		return self._swaptxcreate_bad(['RTC', 'LTC'], exit_val=2, expect1='unrecognized asset')

	def swaptxcreate_bad4(self):
		return self._swaptxcreate_bad(['LTC', 'XTC'], exit_val=2, expect1='unrecognized asset')

	def swaptxcreate_bad5(self):
		return self._swaptxcreate_bad(['LTC'], expect1='USAGE:')

	def swaptxcreate_bad6(self):
		return self._swaptxcreate_bad(['LTC', '1.2345'], expect1='USAGE:')

	def swaptxcreate_bad7(self):
		t = self._swaptxcreate(['BCH', 'LTC'], exit_val=1)
		t.expect('Enter a number> ', '3')
		t.expect('Enter a number> ', '1')
		t.expect('OK? (Y/n): ', 'n')
		return t

	def swaptxcreate_bad8(self):
		addr = make_burn_addr(self.protos[2], mmtype='compressed')
		t = self._swaptxcreate_bad(['BCH', '1.234', addr, 'LTC', 'S'])
		t.expect('to confirm: ', 'NO\n')
		return t

	def swaptxcreate_bad9(self):
		return self._swaptxcreate_bad(['BCH', '1.234', 'S', 'LTC', 'B'], exit_val=2, expect1='invalid command-')

	def swaptxsign1_create(self):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(['LTC', '4.321', f'{self.sid}:S:2', 'BCH', f'{self.sid}:C:2']))

	def swaptxsign1(self):
		return self._swaptxsign()

	def swaptxsend1(self):
		return self._swaptxsend1()

	def swaptxsend1_status(self):
		t = self._swaptxsend1(add_opts=['--status'], spawn_only=True)
		t.expect('in mempool')
		return t

	def swaptxsign2_create(self):
		addr = make_burn_addr(self.protos[2], mmtype='compressed')
		t = self._swaptxcreate(['LTC', '4.56789', f'{self.sid}:S:3', 'BCH', addr])
		t.expect('to confirm: ', 'YES\n') # confirm non-MMGen destination
		return self._swaptxcreate_ui_common(t)

	def swaptxsign2(self):
		return self._swaptxsign(add_opts=['--allow-non-wallet-swap'], expect='swap to non-wallet address')

	def swaptxsend2(self):
		return self._swaptxsend()

	def swaptxbump1(self):
		return self._swaptxbump('20s', add_opts=['--allow-non-wallet-swap'])

	def swaptxbump2(self): # create one-output TX back to self to rescue funds
		return self._swaptxbump('40s', output_args=[f'{self.sid}:S:4'])

	def swaptxsign3(self):
		return self.swaptxsign2()

	def swaptxsend3(self):
		return self._swaptxsend()

	def swaptxsign4(self):
		return self._swaptxsign()

	def swaptxsend4(self):
		return self._swaptxsend()

	def generate0(self):
		return self._generate_for_proto(0)

	def generate1(self):
		return self._generate_for_proto(1)

	def generate2(self):
		return self._generate_for_proto(2)

	def swap_bal1(self):
		return self._bob_bal(1, '10.67894238')

	def swap_bal2(self):
		return self._bob_bal(1, '8.90148152')

	def swap_bal3(self):
		return self._bob_bal(0, '999.99990407')

	def swaptxsign1_do(self):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(['LTC', '1.777444', f'{self.sid}:B:2', 'BCH', f'{self.sid}:C:2'], action='txdo'),
			sign_and_send = True,
			file_desc = 'Sent transaction')

	def swaptxsign_bad1_create(self):
		return self.swaptxcreate6()

	def swaptxsign_bad1(self):
		return self._swaptxsign_bad('non-wallet address forbidden')

	def swaptxsign_bad2_create(self):
		return self.swaptxcreate1(idx=4)

	def swaptxsign_bad2(self):
		return self._swaptxsign()

	def swaptxsend_bad2(self):
		import json
		from mmgen.tx.file import json_dumps
		from mmgen.util import make_chksum_6
		fn = self.get_file_with_ext('sigtx')
		with open(fn) as fh:
			data = json.load(fh)
		data['MMGenTransaction']['swap_quote_expiry'] -= 2400
		data['chksum'] = make_chksum_6(json_dumps(data['MMGenTransaction']))
		with open(fn, 'w') as fh:
			json.dump(data, fh)
		t = self.spawn('mmgen-txsend', ['-d', self.tmpdir, '--bob', fn], exit_val=1)
		t.expect('expired')
		return t

	run_setup_bip39 = CmdTestAutosign.run_setup_bip39
	run_setup = CmdTestAutosign.run_setup

	def swap_wait_loop_start(self):
		return self.wait_loop_start(add_opts=['--allow-non-wallet-swap'])

	def autosign_swaptxcreate1(self):
		return self._user_txcreate(
			'bob',
			progname = 'swaptxcreate',
			ui_handler = self._swaptxcreate_ui_common,
			output_args = ['BTC', '8.88', f'{self.sid}:S:3', 'LTC', f'{self.sid}:S:3'])

	def autosign_swaptxsend1(self):
		return self._user_txsend('bob', need_rbf=True)

	def autosign_swaptxbump1(self):
		return self._user_txcreate(
			'bob',
			progname = 'txbump',
			ui_handler = self._swaptxbump_ui_common_new_outputs,
			output_args = [f'{self.sid}:S:3'])

	def autosign_swaptxsend2(self):
		return self._user_txsend('bob', need_rbf=True)

	# admin methods:

	def sleep(self):
		import time
		time.sleep(1000)
		return 'ok'

	def listaddresses0(self):
		return self._listaddresses(0)

	def listaddresses1(self):
		return self._listaddresses(1)

	def listaddresses2(self):
		return self._listaddresses(2)

	def mempool0(self):
		return self._mempool(0)

	def mempool1(self):
		return self._mempool(1)

	def mempool2(self):
		return self._mempool(2)

	def swap_server_stop(self):
		return self._thornode_server_stop()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.rpc.remote: THORChain base protocol remote RPC client for the MMGen Project
"""

import json

from ....http import HTTPClient
from ....rpc.remote import RemoteRPCClient

# throws exception on error:
def process_response(json_response, errmsg):
	data = json.loads(json_response)
	if data['result'] is None:
		from ....util import die
		die('RPCFailure', errmsg)
	return data['result']

# HTTP POST, JSON-RPC response:
class ThornodeRemoteRPCClient(HTTPClient):

	timeout = 30

	def __init__(self, cfg, proto, *, network_proto=None, host=None):
		for k, v in proto.rpc_remote_rpc_params.items():
			setattr(self, k, v)
		super().__init__(cfg, network_proto=network_proto, host=host)

# HTTP GET, params in query string, JSON-RPC response:
class ThornodeRemoteRESTClient(HTTPClient):

	http_hdrs = {'Content-Type': 'application/json'}
	timeout = 5

	def __init__(self, cfg, proto, *, network_proto=None, host=None):
		for k, v in proto.rpc_remote_rest_params.items():
			setattr(self, k, v)
		super().__init__(cfg, network_proto=network_proto, host=host)

class THORChainRemoteRPCClient(RemoteRPCClient):
	server_proto = 'THORChain'

	def __init__(self, cfg, proto):
		for k, v in proto.rpc_remote_params.items():
			setattr(self, k, v)
		super().__init__(cfg, proto)
		self.caps = ('lbl_id',)
		self.rest_api = ThornodeRemoteRESTClient(cfg, proto)
		self.rpc_api = ThornodeRemoteRPCClient(cfg, proto)

	def get_balance(self, addr, *, block=None):
		res = process_response(
			self.rest_api.get(path=f'/bank/balances/{addr}'),
			errmsg =  f'address ‘{addr}’ not found in blockchain')
		rune_res = [d for d in res if d['denom'] == 'rune']
		assert len(rune_res) == 1, f'{rune_res}: result length is not one!'
		return self.proto.coin_amt(int(rune_res[0]['amount']), from_unit='satoshi')

	def get_account_info(self, addr, *, block=None):
		return process_response(
			self.rest_api.get(path=f'/auth/accounts/{addr}'),
			errmsg =  f'address ‘{addr}’ not found in blockchain')['value']

	def get_tx_info(self, txid):
		return process_response(
			self.rpc_api.post(
				path = '/tx',
				data = {'hash': '0x' + txid}),
			errmsg = f'get info for transaction {txid} failed')

	def tx_op(self, txhex, op=None):
		assert isinstance(txhex, str)
		assert op in ('check_tx', 'broadcast_tx_sync', 'broadcast_tx_async')
		return process_response(
			self.rpc_api.post(
				path = '/' + op,
				data = {'tx': '0x' + txhex}),
			errmsg = f'transaction operation ‘{op}’ failed')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: check that balances of output addrs are zero?

"""
mmgen-split: Split funds after a replayable chain fork using a timelocked transaction
             UNMAINTAINED
"""

from .cfg import Config, gc
from .util import gmsg, die

opts_data = {
	'text': {
		'desc': f"""
               Split funds in an {gc.proj_name} wallet after a chain fork using a
               timelocked transaction
		 """,
		'usage':'[opts] [output addr1] [output addr2]',
		'options': """
-h, --help           Print this help message
--, --longhelp       Print help message for long (global) options
-f, --tx-fees=     f The transaction fees for each chain (comma-separated)
-c, --other-coin=  c The coin symbol of the other chain (default: {oc})
-B, --no-blank       Don't blank screen before displaying unspent outputs
-d, --outdir=      d Specify an alternate directory 'd' for output
-m, --minconf=     n Minimum number of confirmations required to spend
                     outputs (default: 1)
-q, --quiet          Suppress warnings; overwrite files without prompting
-R, --no-rbf         Make transaction non-replaceable (non-replace-by-fee
                     according to BIP 125)
-v, --verbose        Produce more verbose output
-y, --yes            Answer 'yes' to prompts, suppress non-essential output
-R, --rpc-host2=   h Host the other coin daemon is running on (default: none)
-l, --locktime=    t Lock time (block height or unix seconds)
                     (default: {bh})
""",
	'notes': f"""\n
This command creates two transactions: one (with the timelock) to be broadcast
on the long chain and one on the short chain after a replayable chain fork.
Only {gc.proj_name} addresses may be spent to.

The command must be run on the longest chain.  The user is reponsible for
ensuring that the current chain is the longest.  The other chain is specified
on the command line, or it defaults to the most recent replayable fork of the
current chain.

For the split to have a reasonable chance of succeeding, the long chain should
be well ahead of the short one (by more than 20 blocks or so) and transactions
should have a good chance of confirming quickly on both chains.  For this
larger than normal fees may be required.  Fees may be specified on the command
line, or network fee estimation may be used.

If the split fails (i.e. the long-chain TX is broadcast and confirmed on the
short chain), no funds are lost.  A new split attempt can be made with the
long-chain transaction's output as an input for the new split transaction.
This process can be repeated as necessary until the split succeeds.

IMPORTANT: Timelock replay protection offers NO PROTECTION against reorg
attacks on the majority chain or reorg attacks on the minority chain if the
minority chain is ahead of the timelock.  If the reorg'd minority chain is
behind the timelock, protection is contingent on getting the non-timelocked
transaction reconfirmed before the timelock expires. Use at your own risk.
"""
	},
	'code': {
		'options': lambda proto, s: s.format(
			oc = proto.forks[-1][2].upper(),
			bh = 'current block height'),
	}
}

cfg = Config(opts_data=opts_data, need_amt=False)

proto = cfg._proto

die(1, 'This command is disabled')

# the following code is broken:
cfg.other_coin = cfg.other_coin.upper() if cfg.other_coin else proto.forks[-1][2].upper()
if cfg.other_coin.lower() not in [e[2] for e in proto.forks if e[3] is True]:
	die(1, f'{cfg.other_coin!r}: not a replayable fork of {proto.coin} chain')

if len(cfg._args) != 2:
	die(1, f'This command requires exactly two {gc.proj_name} addresses as arguments')

from .addr import MMGenID
try:
	mmids = [MMGenID(proto, a) for a in cfg._args]
except:
	die(1, 'Command line arguments must be valid MMGen IDs')

if mmids[0] == mmids[1]:
	die(2, f'Both transactions have the same output! ({mmids[0]})')

from .tx import MMGenSplitTX
from .protocol import init_proto

if cfg.tx_fees:
	for idx, g_coin in ((1, cfg.other_coin), (0, proto.coin)):
		proto = init_proto(cfg, g_coin)
		cfg.fee = cfg.tx_fees.split(',')[idx]
#		opts.opt_is_tx_fee('foo', cfg.fee, 'transaction fee') # raises exception on error

tx1 = MMGenSplitTX(proto)
cfg.no_blank = True

async def main():
	gmsg(f'Creating timelocked transaction for long chain ({proto.coin})')
	locktime = int(cfg.locktime)
	if not locktime:
		from .rpc import rpc_init
		rpc = rpc_init(proto)
		locktime = rpc.call('getblockcount')
	tx1.create(mmids[0], locktime)

	tx1.format()
	tx1.create_fn()

	gmsg(f'\nCreating transaction for short chain ({cfg.other_coin})')

	proto2 = init_proto(cfg, cfg.other_coin)

	tx2 = MMGenSplitTX(proto2)
	tx2.inputs = tx1.inputs
	tx2.inputs.convert_coin()

	tx2.create_split(mmids[1])

	for tx, desc in ((tx1, 'Long chain (timelocked)'), (tx2, 'Short chain')):
		tx.desc = desc + ' transaction'
		tx.file.write(ask_write=False, ask_overwrite=not cfg.yes, ask_write_default_yes=False)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tx.file: Transaction file operations for the MMGen suite
"""

import os, json

from ..util import ymsg, make_chksum_6, die
from ..obj import MMGenObject, HexStr, MMGenTxID, CoinTxID, MMGenTxComment

class txdata_json_encoder(json.JSONEncoder):
	def default(self, o):
		if type(o).__name__.endswith('Amt'):
			return str(o)
		elif type(o).__name__ == 'OpReturnData':
			return repr(o)
		else:
			return json.JSONEncoder.default(self, o)

def json_dumps(data):
	return json.dumps(data, separators = (',', ':'), cls=txdata_json_encoder)

def get_proto_from_coin_id(tx, coin_id, chain):
	coin, tokensym = coin_id.split(':') if ':' in coin_id else (coin_id, None)

	from ..protocol import CoinProtocol, init_proto
	network = CoinProtocol.Base.chain_name_to_network(tx.cfg, coin, chain)

	return init_proto(tx.cfg, coin, network=network, need_amt=True, tokensym=tokensym)

def eval_io_data(tx, data, *, desc):
	if not (desc == 'outputs' and tx.proto.base_coin == 'ETH'): # ETH txs can have no outputs
		assert len(data), f'no {desc}!'
	for d in data:
		d['amt'] = tx.proto.coin_amt(d['amt'])
	io, io_list = {
		'inputs':  (tx.Input, tx.InputList),
		'outputs': (tx.Output, tx.OutputList),
	}[desc]
	return io_list(parent=tx, data=[io(tx.proto, **d) for d in data])

class MMGenTxFile(MMGenObject):
	data_label = 'MMGenTransaction'
	attrs = {
		'chain': None,
		'txid': MMGenTxID,
		'send_amt': 'skip',
		'timestamp': None,
		'blockcount': None,
		'serialized': None}
	extra_attrs = {
		'locktime': None,
		'comment': MMGenTxComment,
		'coin_txid': CoinTxID,
		'sent_timestamp': None,
		'is_swap': None}

	def __init__(self, tx):
		self.tx       = tx
		self.fmt_data = None
		self.filename = None

	def parse(self, infile, *, metadata_only=False, quiet_open=False):
		tx = self.tx
		from ..fileutil import get_data_from_file
		data = get_data_from_file(tx.cfg, infile, desc=f'{tx.desc} data', quiet=quiet_open)
		if len(data) > tx.cfg.max_tx_file_size:
			die('MaxFileSizeExceeded',
				f'Transaction file size exceeds limit ({tx.cfg.max_tx_file_size} bytes)')
		return (self.parse_data_json if data[0] == '{' else self.parse_data_legacy)(data, metadata_only)

	def parse_data_json(self, data, metadata_only):
		tx = self.tx
		tx.file_format = 'json'
		outer_data = json.loads(data)
		data = outer_data[self.data_label]
		if outer_data['chksum'] != make_chksum_6(json_dumps(data)):
			chk = make_chksum_6(json_dumps(data))
			die(3, f'{self.data_label}: invalid checksum for TxID {data["txid"]} ({chk} != {outer_data["chksum"]})')

		tx.proto = get_proto_from_coin_id(tx, data['coin_id'], data['chain'])

		for k, v in self.attrs.items():
			if v != 'skip':
				setattr(tx, k, v(data[k]) if v else data[k])

		if metadata_only:
			return

		for k, v in self.extra_attrs.items():
			if k in data:
				setattr(tx, k, v(data[k]) if v else data[k])

		if tx.is_swap:
			for k, v in tx.swap_attrs.items():
				if k in data:
					setattr(tx, k, v(data[k]) if v else data[k])

		for k in ('inputs', 'outputs'):
			setattr(tx, k, eval_io_data(tx, data[k], desc=k))

		tx.check_txfile_hex_data()

		tx.parse_txfile_serialized_data() # Ethereum RLP or JSON data

		assert tx.proto.coin_amt(data['send_amt']) == tx.send_amt, f'{data["send_amt"]} != {tx.send_amt}'

	def parse_data_legacy(self, data, metadata_only):
		tx = self.tx
		tx.file_format = 'legacy'

		def deserialize(raw_data, *, desc):
			from ast import literal_eval
			try:
				return literal_eval(raw_data)
			except:
				if desc == 'inputs':
					ymsg('Warning: transaction data appears to be in old format')
				import re
				return literal_eval(re.sub(r"[A-Za-z]+?\(('.+?')\)", r'\1', raw_data))

		desc = 'data'
		try:
			tx_data = data.splitlines()
			assert len(tx_data) >= 5, 'number of lines less than 5'
			assert len(tx_data[0]) == 6, 'invalid length of first line'
			assert HexStr(tx_data.pop(0)) == make_chksum_6(' '.join(tx_data)), 'file data does not match checksum'

			if len(tx_data) == 7:
				desc = 'sent timestamp'
				(_, tx.sent_timestamp) = tx_data.pop(-1).split()
				assert _ == 'Sent', 'invalid sent timestamp line'

			if len(tx_data) == 6:
				assert len(tx_data[-1]) == 64, 'invalid coin TxID length'
				desc = 'coin TxID'
				tx.coin_txid = CoinTxID(tx_data.pop(-1))

			if len(tx_data) == 5:
				# rough check: allow for 4-byte utf8 characters + base58 (4 * 11 / 8 = 6 (rounded up))
				assert len(tx_data[-1]) < MMGenTxComment.max_len*6, 'invalid comment length'
				c = tx_data.pop(-1)
				if c != '-':
					desc = 'encoded comment (not base58)'
					from ..baseconv import baseconv
					comment = baseconv('b58').tobytes(c).decode()
					assert comment is not False, 'invalid comment'
					desc = 'comment'
					tx.comment = MMGenTxComment(comment)

			desc = 'number of lines' # four required lines
			io_data = {}
			(metadata, tx.serialized, io_data['inputs'], io_data['outputs']) = tx_data
			assert len(metadata) < 100, 'invalid metadata length' # rough check
			metadata = metadata.split()

			if metadata[-1].startswith('LT='):
				desc = 'locktime'
				tx.locktime = int(metadata.pop()[3:])

			desc = 'coin token in metadata'
			coin_id = metadata.pop(0) if len(metadata) == 6 else 'BTC'

			desc = 'chain token in metadata'
			tx.chain = metadata.pop(0).lower() if len(metadata) == 5 else 'mainnet'

			desc = 'coin_id or chain'
			tx.proto = get_proto_from_coin_id(tx, coin_id, tx.chain)

			desc = 'metadata (4 items)'
			(txid, send_amt, tx.timestamp, blockcount) = metadata

			desc = 'TxID in metadata'
			tx.txid = MMGenTxID(txid)
			desc = 'block count in metadata'
			tx.blockcount = int(blockcount)

			if metadata_only:
				return

			desc = 'transaction file hex data'
			tx.check_txfile_hex_data()
			desc = 'Ethereum RLP or JSON data'
			tx.parse_txfile_serialized_data()
			for k in ('inputs', 'outputs'):
				desc = f'{k} data'
				res = deserialize(io_data[k], desc=k)
				for d in res:
					if 'label' in d:
						d['comment'] = d['label']
						del d['label']
				setattr(tx, k, eval_io_data(tx, res, desc=k))
			desc = 'send amount in metadata'
			assert tx.proto.coin_amt(send_amt) == tx.send_amt, f'{send_amt} != {tx.send_amt}'
		except Exception as e:
			die(2, f'Invalid {desc} in transaction file: {e!s}')

	def make_filename(self):
		tx = self.tx
		def gen_filename():
			yield tx.txid
			if tx.coin != 'BTC':
				yield '-' + tx.dcoin
			yield f'[{tx.send_amt!s}'
			if tx.is_replaceable():
				yield ',{}'.format(tx.fee_abs2rel(tx.fee, to_unit=tx.fn_fee_unit))
			if tx.get_serialized_locktime():
				yield f',tl={tx.get_serialized_locktime()}'
			yield ']'
			if tx.proto.testnet:
				yield '.' + tx.proto.network
			yield '.' + tx.ext
		return ''.join(gen_filename())

	def format(self):
		tx = self.tx
		coin_id = tx.coin + ('' if tx.coin == tx.dcoin else ':'+tx.dcoin)

		def format_data_legacy():

			def amt_to_str(d):
				return {k: (str(d[k]) if k == 'amt' else d[k]) for k in d}

			lines = [
				'{}{} {} {} {} {}{}'.format(
					(f'{coin_id} ' if coin_id and tx.coin != 'BTC' else ''),
					tx.chain.upper(),
					tx.txid,
					tx.send_amt,
					tx.timestamp,
					tx.blockcount,
					(f' LT={tx.locktime}' if tx.locktime else ''),
				),
				tx.serialized,
				ascii([amt_to_str(e._asdict()) for e in tx.inputs]),
				ascii([amt_to_str(e._asdict()) for e in tx.outputs])
			]

			if tx.comment:
				from ..baseconv import baseconv
				lines.append(baseconv('b58').frombytes(tx.comment.encode(), tostr=True))

			if tx.coin_txid:
				if not tx.comment:
					lines.append('-') # keep old tx files backwards compatible
				lines.append(tx.coin_txid)

			if tx.sent_timestamp:
				lines.append(f'Sent {tx.sent_timestamp}')

			return '\n'.join([make_chksum_6(' '.join(lines))] + lines) + '\n'

		def format_data_json():
			data = json_dumps({
					'coin_id': coin_id
				} | {
					k: getattr(tx, k) for k in self.attrs
				} | {
					'inputs':  [e._asdict() for e in tx.inputs],
					'outputs': [{k:v for k,v in e._asdict().items() if not (type(v) is bool and v is False)}
									for e in tx.outputs]
				} | {
					k: getattr(tx, k) for k in self.extra_attrs if getattr(tx, k)
				} | ({
					k: getattr(tx, k) for k in tx.swap_attrs if getattr(tx, k, None)
				} if tx.is_swap else {}))
			return '{{"{}":{},"chksum":"{}"}}'.format(self.data_label, data, make_chksum_6(data))

		fmt_data = {'json': format_data_json, 'legacy': format_data_legacy}[tx.file_format]()

		if len(fmt_data) > tx.cfg.max_tx_file_size:
			die('MaxFileSizeExceeded', f'Transaction file size exceeds limit ({tx.cfg.max_tx_file_size} bytes)')

		return fmt_data

	def write(self, *,
		add_desc              = '',
		outdir                = None,
		ask_write             = True,
		ask_write_default_yes = False,
		ask_tty               = True,
		ask_overwrite         = True):

		if ask_write is False:
			ask_write_default_yes = True

		if not self.filename:
			self.filename = self.make_filename()

		if not self.fmt_data:
			self.fmt_data = self.format()

		from ..fileutil import write_data_to_file
		write_data_to_file(
			cfg                   = self.tx.cfg,
			outfile               = os.path.join((outdir or ''), self.filename),
			data                  = self.fmt_data,
			desc                  = self.tx.desc + add_desc,
			ask_overwrite         = ask_overwrite,
			ask_write             = ask_write,
			ask_tty               = ask_tty,
			ask_write_default_yes = ask_write_default_yes,
			ignore_opt_outdir     = outdir)

	@classmethod
	def get_proto(cls, cfg, filename, *, quiet_open=False):
		from . import BaseTX
		tmp_tx = BaseTX(cfg=cfg)
		cls(tmp_tx).parse(filename, metadata_only=True, quiet_open=quiet_open)
		return tmp_tx.proto



#!/usr/bin/env python3

"""
test.modtest_d.addrparse: address parsing tests for the MMGen suite
"""

from mmgen.color import yellow, cyan
from mmgen.util import msg, msg_r, pp_fmt

from ..include.common import cfg, vmsg

vectors = {
	'btc_mainnet': [
		{'std': '1C5VPtgq9xQ6AcTgMAR3J6GDrs72HC4pS1'},
		{'std': '3AhjTiWHhVJAi1s5CfKMcLzYps12x3gZhg'},
		{'std': 'bc1q6pqnfwwakuuejpm9w52ds342f9d5u36v0qnz7c'}
	],
	'ltc_mainnet': [
		{'std': 'LUbHQNYoy23RByq4dKQotLA4ugk9FhpAMT'},
		{'std': 'MCoZrHYPqYKqvpiwyzzqf3EPxF5no6puEf'},
		{'std': 'ltc1qvmqas4maw7lg9clqu6kqu9zq9cluvllnst5pxs'}
	],
	'xmr_mainnet': [
		{ # xmrseed.vectors[0]:
		'std': '42ey1afDFnn4886T7196doS9GPMzexD9gXpsZJDwVjeRVdFCSoHnv7KPbBeGpzJBzHRCAs9UxqeoyFQMYbqSWYTfJJQAWDm',
		# https://github.com/monero-project/monero/tests/functional_tests/integrated_address.py
		'int': '4CMe2PUhs4J4886T7196doS9GPMzexD9gXpsZJDwVjeRVdFCSoHnv7KPbBeGpzJBzHRCAs9UxqeoyFQMYbqSWYTfSbLRB61BQVATzerHGj',
		'id':  '0123456789abcdef'
		}, {
		'std': '46r4nYSevkfBUMhuykdK3gQ98XDqDTYW1hNLaXNvjpsJaSbNtdXh1sKMsdVgqkaihChAzEy29zEDPMR3NHQvGoZCLGwTerK',
		'int': '4GYjoMG9Y2BBUMhuykdK3gQ98XDqDTYW1hNLaXNvjpsJaSbNtdXh1sKMsdVgqkaihChAzEy29zEDPMR3NHQvGoZCVSs1ZojwrDCGS5rUuo',
		'id':  '1122334455667788'
		}
	],
	'zec_mainnet': [
		{'std': 't1KQYLBvjpmcQuATommo6gx2QTQDLPikB8Q'},
		{'std': 'zceQDpyNwek7dKqF5ZuFGj7YrNVxh7X1aPkrVxDLVxWSiZAFDEuy5C7XNV8VhyZ3ghTPQ61xjCGiyLT3wqpiN1Yi6mdmaCq'},
	],
	'eth_mainnet': [
		{'std': '7e5f4552091a69125d5dfcb7b8c2659029395bdf'},
	],
}

def test_network(proto, addrs):

	def check_equal(a, b):
		assert a == b, f'{a.hex()} != {b.hex()}'

	def check_bytes(addr):
		if addr.parsed.ver_bytes is not None:
			check_equal(
				addr.parsed.ver_bytes,
				proto.addr_fmt_to_ver_bytes.get(addr.addr_fmt))
		check_equal(
			addr.parsed.data + ((addr.parsed.payment_id or b'') if proto.coin == 'XMR' else b''),
			addr.bytes)

	def fmt_addr_data(addr):
		return pp_fmt({k:(v.hex() if isinstance(v, bytes) else v) for k, v in addr.parsed._asdict().items()})

	def print_info(addr):
		vmsg('\n{}\n{}\n{}'.format(yellow(addr.addr_fmt), cyan(addr), fmt_addr_data(addr)))

	msg_r(f'Testing {proto.coin} address parsing...')
	vmsg('')

	from mmgen.addr import CoinAddr

	for addr in addrs:
		a1 = CoinAddr(proto, addr['std'])
		print_info(a1)
		check_bytes(a1)
		assert not hasattr(a1.parsed, 'payment_id') or a1.parsed.payment_id is None

		if 'int' in addr:
			a2 = CoinAddr(proto, addr['int'])
			print_info(a2)
			check_bytes(a2)
			check_equal(a1.parsed.data, a2.parsed.data)
			check_equal(a2.parsed.payment_id, bytes.fromhex(addr['id']))

	msg('OK')
	vmsg('')


class unit_test:

	def run_test(self, name, ut):

		from mmgen.protocol import init_proto

		for net_id, addrs in vectors.items():
			coin, network = net_id.split('_')
			if cfg.no_altcoin_deps and coin != 'btc':
				continue
			test_network(
				init_proto(cfg, coin, network=network),
				addrs)

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.daemon: Ethereum base protocol daemon classes
"""

import os

from ...cfg import gc
from ...util import list_gen, get_subclasses
from ...daemon import CoinDaemon, RPCDaemon, _nw, _dd

class ethereum_daemon(CoinDaemon):
	chain_subdirs = _nw('ethereum', 'goerli', 'DevelopmentChain')
	base_rpc_port = 8545  # same for all networks!
	base_authrpc_port = 8551 # same for all networks!
	base_p2p_port = 30303 # same for all networks!
	daemon_port_offset = 100
	network_port_offsets = _nw(0, 10, 20)

	def __init__(self, *args, test_suite=False, **kwargs):

		if not hasattr(self, 'all_daemons'):
			ethereum_daemon.all_daemons = get_subclasses(ethereum_daemon, names=True)

		daemon_idx_offset = (
			self.all_daemons.index(self.id+'_daemon') * self.daemon_port_offset
			if test_suite else 0)

		self.port_offset = daemon_idx_offset + getattr(self.network_port_offsets, self.network)

		super().__init__(*args, test_suite=test_suite, **kwargs)

	def get_rpc_port(self):
		return self.base_rpc_port + self.port_offset

	@property
	def authrpc_port(self):
		return self.base_authrpc_port + self.port_offset

	def get_p2p_port(self):
		return self.base_p2p_port + self.port_offset

	def init_datadir(self):
		self.logdir = super().init_datadir()
		return os.path.join(
			self.logdir,
			self.id,
			getattr(self.chain_subdirs, self.network))

class openethereum_daemon(ethereum_daemon):
	daemon_data = _dd('OpenEthereum', 3003005, '3.3.5')
	version_pat = r'OpenEthereum//v(\d+)\.(\d+)\.(\d+)'
	exec_fn = 'openethereum'
	cfg_file = 'parity.conf'
	datadirs = {
		'linux': [gc.home_dir, '.local', 'share', 'io.parity.ethereum'],
		'darwin': [gc.home_dir, 'Library', 'Application Support', 'io.parity.ethereum'],
		'win32': [os.getenv('LOCALAPPDATA'), 'Parity', 'Ethereum']
	}

	def init_subclass(self):

		self.use_pidfile = self.platform == 'linux' and not self.opt.no_daemonize
		self.use_threads = self.platform in ('win32', 'darwin')

		self.coind_args = list_gen(
			['--no-ws'],
			['--no-ipc'],
			['--no-secretstore'],
			[f'--jsonrpc-port={self.rpc_port}'],
			[f'--port={self.p2p_port}', self.p2p_port],
			[f'--base-path={self.datadir}', self.non_dfl_datadir],
			[f'--chain={self.proto.chain_name}', self.network!='regtest'],
			['--config=dev', self.network=='regtest'], # no presets for mainnet or testnet
			['--mode=offline', self.test_suite or self.network=='regtest'],
			[f'--log-file={self.logfile}', self.non_dfl_datadir],
			['daemon', self.use_pidfile],
			[self.pidfile, self.use_pidfile],
		)

class parity_daemon(openethereum_daemon):
	daemon_data = _dd('Parity', 2007002, '2.7.2')
	version_pat = r'Parity-Ethereum//v(\d+)\.(\d+)\.(\d+)'
	exec_fn = 'parity'

class geth_daemon(ethereum_daemon):
	# v1.14.0 -> ? (v1.15.11 and later OK)
	#   mempool deadlock in dev mode: "transaction indexing is in progress"
	#   https://github.com/ethereum/go-ethereum/issues/29475
	#   offending commit (via git bisect): 0a2f33946b95989e8ce36e72a88138adceab6a23
	daemon_data = _dd('Geth', 1016003, '1.16.3')
	version_pat = r'Geth/v(\d+)\.(\d+)\.(\d+)'
	exec_fn = 'geth'
	use_pidfile = False
	use_threads = True
	avail_opts = ('no_daemonize', 'online')
	version_info_arg = 'version'
	datadirs = {
		'linux': [gc.home_dir, '.ethereum', 'geth'],
		'darwin': [gc.home_dir, 'Library', 'Ethereum', 'geth'],
		'win32': [os.getenv('LOCALAPPDATA'), 'Geth'] # FIXME
	}

	def init_subclass(self):

		self.coind_args = list_gen(
			['node', self.id == 'reth'],
			['--quiet', self.id == 'reth'],
			['--disable-dns-discovery', self.id == 'reth' and self.test_suite],
			['--verbosity=0', self.id == 'geth'],
			['--ipcdisable'], # IPC-RPC: if path to socket is longer than 108 chars, geth fails to start
			['--http'],
			['--http.api=eth,web3,txpool'],
			[f'--http.port={self.rpc_port}'],
			[f'--authrpc.port={self.authrpc_port}'],
			[f'--port={self.p2p_port}', self.p2p_port], # geth binds p2p port even with --maxpeers=0
			[f'--discovery.port={self.p2p_port}', self.id == 'reth' and self.p2p_port],
			['--maxpeers=0', self.id == 'geth' and not self.opt.online],
			[f'--datadir={self.datadir}', self.non_dfl_datadir],
			['--holesky', self.network=='testnet' and self.id == 'geth'],
			['--chain=holesky', self.network=='testnet' and self.id == 'reth'],
			['--dev', self.network=='regtest'],
		)

class reth_daemon(geth_daemon):
	daemon_data = _dd('Reth', 1007000, '1.7.0')
	version_pat = r'reth/v(\d+)\.(\d+)\.(\d+)'
	exec_fn = 'reth'
	version_info_arg = '--version'
	datadirs = {
		'linux': [gc.home_dir, '.local', 'share', 'reth'],
	}

# https://github.com/ledgerwatch/erigon
class erigon_daemon(geth_daemon):
	daemon_data = _dd('Erigon', 2022099099, '2022.99.99')
	version_pat = r'erigon/(\d+)\.(\d+)\.(\d+)'
	exec_fn = 'erigon'
	private_ports = _nw(9090, 9091, 9092) # testnet and regtest are non-standard
	torrent_ports = _nw(42069, 42070, None) # testnet is non-standard
	version_info_arg = '--version'
	datadirs = {
		'linux': [gc.home_dir, '.local', 'share', 'erigon'],
		'win32': [os.getenv('LOCALAPPDATA'), 'Erigon'] # FIXME
	}

	def init_subclass(self):

		if self.network == 'regtest':
			self.force_kill = True

		self.coind_args = list_gen(
			['--verbosity=0'],
			[f'--port={self.p2p_port}', self.p2p_port],
			['--maxpeers=0', not self.opt.online],
			[f'--private.api.addr=127.0.0.1:{self.private_port}'],
			[f'--datadir={self.datadir}', self.non_dfl_datadir],
			['--chain=goerli', self.network=='testnet'],
			[f'--torrent.port={self.torrent_ports.testnet}', self.network=='testnet'],
			['--chain=dev', self.network=='regtest'],
			['--mine', self.network=='regtest'],
		)

		self.rpc_d = erigon_rpcdaemon(
			cfg          = self.cfg,
			proto        = self.proto,
			rpc_port     = self.rpc_port,
			private_port = self.private_port,
			test_suite   = self.test_suite,
			datadir      = self.datadir)

	def start(self, *, quiet=False, silent=False):
		super().start(quiet=quiet, silent=silent)
		self.rpc_d.debug = self.debug
		return self.rpc_d.start(quiet=quiet, silent=silent)

	def stop(self, *, quiet=False, silent=False):
		self.rpc_d.debug = self.debug
		self.rpc_d.stop(quiet=quiet, silent=silent)
		return super().stop(quiet=quiet, silent=silent)

	@property
	def start_cmds(self):
		return [self.start_cmd, self.rpc_d.start_cmd]

class erigon_rpcdaemon(RPCDaemon):

	master_daemon = 'erigon_daemon'
	rpc_desc = 'Erigon'
	exec_fn = 'rpcdaemon'
	use_pidfile = False
	use_threads = True

	def __init__(self, cfg, proto, *, rpc_port, private_port, test_suite, datadir):

		self.proto = proto
		self.test_suite = test_suite

		super().__init__(cfg)

		self.network = proto.network
		self.rpc_port = rpc_port
		self.datadir = datadir

		self.daemon_args = list_gen(
			['--verbosity=0'],
			[f'--private.api.addr=127.0.0.1:{private_port}'],
			[f'--http.port={self.rpc_port}'],
			[f'--datadir={self.datadir}'],
			['--http.api=eth,erigon,web3,net,debug,trace,txpool,parity'],
		)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.json: export and import tracking wallet to JSON format
"""

from collections import namedtuple
from ....tw.json import TwJSON
from ....tw.shared import TwMMGenID

class BitcoinTwJSON(TwJSON):

	class Base(TwJSON.Base):

		can_prune = True

		@property
		def mappings_json(self):
			return self.json_dump([(e.mmgen_id, e.address) for e in self.entries])

		@property
		def num_entries(self):
			return len(self.entries)

	class Import(TwJSON.Import, Base):

		info_msg = """
			This utility will create a new tracking wallet, import the addresses from
			the JSON dump into it and update their balances. The operation may take a
			few minutes.
		"""

		blockchain_rescan_warning = """
			Balances have been updated in the new tracking wallet.  However, the wallet
			is unaware of the used state of any spent addresses without balances, which
			creates the danger of address reuse, especially when automatic change address
			selection is in effect.

			To avoid this danger and restore full tracking wallet functionality, rescan
			the blockchain for used addresses by running ‘mmgen-tool rescan_blockchain’.
		"""

		@property
		async def tracking_wallet_exists(self):
			return await self.twctl.rpc.tracking_wallet_exists

		async def create_tracking_wallet(self):
			try:
				await self.twctl.rpc.check_or_create_daemon_wallet()
				return True
			except:
				return False

		async def get_entries(self):
			entries_in = [self.entry_tuple_in(*e) for e in self.data['data']['entries']]
			return sorted(
				[self.entry_tuple(
					TwMMGenID(self.proto, d.mmgen_id),
					d.address,
					getattr(d, 'amount', None),
					d.comment)
						for d in entries_in],
				key = lambda x: x.mmgen_id.sort_key)

		async def do_import(self, batch):
			import_tuple = namedtuple('import_data', ['addr', 'twmmid', 'comment'])
			await self.twctl.import_address_common(
				[import_tuple(e.address, e.mmgen_id, e.comment) for e in self.entries],
				batch = batch)
			return [e.address for e in self.entries]

	class Export(TwJSON.Export, Base):

		@property
		async def addrlist(self):
			if not hasattr(self, '_addrlist'):
				if self.prune:
					from .prune import TwAddressesPrune
					self._addrlist = al = await TwAddressesPrune(
						self.cfg,
						self.proto,
						get_data  = True,
						warn_used = self.warn_used)
					await al.view_filter_and_sort()
					self.pruned = al.do_prune()
				else:
					from .addresses import TwAddresses
					self._addrlist = await TwAddresses(self.cfg, self.proto, get_data=True)
			return self._addrlist

		async def get_entries(self): # TODO: include 'received' field
			return sorted(
				[self.entry_tuple(d.twmmid, d.addr, d.amt, d.comment)
					for d in (await self.addrlist).data],
				key = lambda x: x.mmgen_id.sort_key)

		@property
		async def entries_out(self):
			return [[getattr(d, k) for k in self.keys] for d in self.entries]

		@property
		async def total(self):
			return (await self.addrlist).total



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_d.opts: options processing tests for the MMGen cmdtest.py test suite
"""

import os, time

from .base import CmdTestBase

class CmdTestOpts(CmdTestBase):
	'command-line options parsing and processing'
	networks = ('btc',)
	tmpdir_nums = [41]
	cmd_group = (
		('opt_helpscreen',       (41, 'helpscreen output', [])),
		('opt_noargs',           (41, 'invocation with no user options or arguments', [])),
		('opt_good1',            (41, 'good opts (long opts only)', [])),
		('opt_good2',            (41, 'good opts (mixed short and long opts)', [])),
		('opt_good3',            (41, 'good opts (max arg count)', [])),
		('opt_good4',            (41, 'good opts (maxlen arg)', [])),
		('opt_good5',            (41, 'good opts (long opt substring)', [])),
		('opt_good6',            (41, 'good global opt (--coin=xmr)', [])),
		('opt_good7',            (41, 'good global opt (--coin xmr)', [])),
		('opt_good8',            (41, 'good global opt (--pager)', [])),
		('opt_good9',            (41, 'good cmdline arg ‘-’', [])),
		('opt_good10',           (41, 'good cmdline arg ‘-’ with arg', [])),
		('opt_good11',           (41, 'good cmdline arg ‘-’ with option', [])),
		('opt_good12',           (41, 'good cmdline opt (short opt with option)', [])),
		('opt_good13',           (41, 'good cmdline opt (short opt with option)', [])),
		('opt_good14',           (41, 'good cmdline opt (combined short opt with option)', [])),
		('opt_good15',           (41, 'good cmdline opt (combined short opt with option)', [])),
		('opt_good16',           (41, 'good cmdline opt (param with equals signs)', [])),
		('opt_good17',           (41, 'good cmdline opt (param with equals signs)', [])),
		('opt_good18',           (41, 'good cmdline opt (param with equals signs)', [])),
		('opt_good19',           (41, 'good cmdline opt (param with equals signs)', [])),
		('opt_good20',           (41, 'good cmdline opt (opt + negated opt)', [])),
		('opt_good21',           (41, 'good cmdline opt (negated negative opt)', [])),
		('opt_good22',           (41, 'good cmdline opt (opt + negated opt [substring])', [])),
		('opt_good23',           (41, 'good cmdline opt (negated negative opt [substring])', [])),
		('opt_good24',           (41, 'good cmdline opt (negated opt + opt [substring])', [])),
		('opt_good25',           (41, 'good cmdline opt (--btc-rpc-host)', [])),
		('opt_good26',           (41, 'good cmdline opt (--btc-rpc-port)', [])),
		('opt_good27',           (41, 'good cmdline opt (--btc-ignore-daemon-version)', [])),
		('opt_good28',           (41, 'good cmdline opt (--bch-cashaddr)', [])),
		('opt_good29',           (41, 'good cmdline opt (--etc-max-tx-fee=0.1)', [])),
		('opt_good30',           (41, 'good cmdline opt (--eth-chain-names=foo,bar)', [])),
		('opt_good31',           (41, 'good cmdline opt (--xmr-rpc-port=28081)', [])),
		('opt_bad_param',        (41, 'bad global opt (--pager=1)', [])),
		('opt_bad_infile',       (41, 'bad infile parameter', [])),
		('opt_bad_outdir',       (41, 'bad outdir parameter', [])),
		('opt_bad_incompatible', (41, 'incompatible opts', [])),
		('opt_bad_autoset',      (41, 'invalid autoset value', [])),
		('opt_invalid_1',        (41, 'invalid cmdline opt ‘--x’', [])),
		('opt_invalid_2',        (41, 'invalid cmdline opt ‘---’', [])),
		('opt_invalid_5',        (41, 'invalid cmdline opt (missing parameter)', [])),
		('opt_invalid_6',        (41, 'invalid cmdline opt (missing parameter)', [])),
		('opt_invalid_7',        (41, 'invalid cmdline opt (parameter not required)', [])),
		('opt_invalid_8',        (41, 'invalid cmdline opt (non-existent option)', [])),
		('opt_invalid_9',        (41, 'invalid cmdline opt (non-existent option)', [])),
		('opt_invalid_10',       (41, 'invalid cmdline opt (missing parameter)', [])),
		('opt_invalid_11',       (41, 'invalid cmdline opt (missing parameter)', [])),
		('opt_invalid_12',       (41, 'invalid cmdline opt (non-existent option)', [])),
		('opt_invalid_13',       (41, 'invalid cmdline opt (ambiguous long opt substring)', [])),
		('opt_invalid_14',       (41, 'invalid cmdline opt (long opt substring too short)', [])),
		('opt_invalid_15',       (41, 'invalid cmdline (too many args)', [])),
		('opt_invalid_16',       (41, 'invalid cmdline (overlong arg)', [])),
		('opt_invalid_17',       (41, 'invalid cmdline opt (--btc-rpc-host without ‘need_proto’)', [])),
		('opt_invalid_18',       (41, 'invalid cmdline opt (--btc-rpc-port without ‘need_proto’)', [])),
		('opt_invalid_19',       (41, 'invalid cmdline opt (--btc-rpc-port with non-integer param)', [])),
		('opt_invalid_21',       (41, 'invalid cmdline opt (--btc-foo)', [])),
		('opt_invalid_22',       (41, 'invalid cmdline opt (--btc-rpc-host with missing param)', [])),
		('opt_invalid_23',       (41, 'invalid cmdline opt (--btc-ignore-daemon-version with param)', [])),
		('opt_invalid_24',       (41, 'invalid cmdline opt (--bch-cashaddr without ‘need_proto’)', [])),
		('opt_invalid_25',       (41, 'invalid cmdline opt (--bch-cashaddr without parameter)', [])),
		('opt_invalid_26',       (41, 'invalid cmdline opt (--bch-cashaddr with non-bool parameter)', [])),
		('opt_invalid_27',       (41, 'invalid cmdline opt (--ltc-cashaddr)', [])),
		('opt_invalid_28',       (41, 'invalid cmdline opt (--xmr-max-tx-fee)', [])),
		('opt_invalid_29',       (41, 'invalid cmdline opt (--eth-max-tx-fee without parameter)', [])),
		('opt_invalid_30',       (41, 'invalid cmdline opt (--eth-max-tx-fee with non-numeric parameter)', [])),
		('opt_invalid_31',       (41, 'invalid cmdline opt (--bch-cashaddr without --coin=bch)', [])),
		('opt_invalid_32',       (41, 'invalid cmdline opt (--eth-chain-names without --coin=eth)', [])),
		('opt_invalid_33',       (41, 'invalid cmdline opt (--xmr-rpc-host)', [])),
		('opt_invalid_34',       (41, 'invalid cmdline opt (--eth-rpc-user)', [])),
	)

	def spawn_prog(self, args, opts=[], exit_val=None, need_proto=False):
		return self.spawn(
			'test/misc/opts.py',
			opts + args,
			cmd_dir  = '.',
			exit_val = exit_val,
			env      = {'TEST_MISC_OPTS_NEEDS_PROTO': '1' if need_proto else ''})

	def check_vals(self, args, vals, check=True, need_proto=False):
		show_opts = [a.removeprefix('cfg.') for a, b in vals if a.startswith('cfg.')]
		t = self.spawn_prog(
			args,
			opts       = ['--show-opts=' + ','.join(show_opts)] if show_opts else [],
			need_proto = need_proto)
		if check:
			for k, v in vals:
				t.expect(rf'{k}:\s+{v}', regex=True)
		return t

	def do_run(self, args, expect, exit_val, regex=False):
		t = self.spawn_prog(args, exit_val=exit_val or None)
		t.expect(expect, regex=regex)
		return t

	def opt_helpscreen(self):
		expect = r'OPTS.PY: Opts test.*USAGE:\s+opts.py'
		if not self.cfg.pexpect_spawn:
			expect += r'.*--minconf.*NOTES FOR THIS.*a note'
		t = self.do_run(['--help'], expect, 0, regex=True)
		if t.pexpect_spawn:
			time.sleep(0.4)
			t.send('q')
		return t

	def opt_noargs(self):
		return self.check_vals(
			[],
			(
				('cfg.foo',                 'None'),         # added opt
				('cfg.print_checksum',      'None'),         # sets 'quiet'
				('cfg.quiet',               'False'),        # _incompatible_opts
				('cfg.verbose',             'False'),        # _incompatible_opts
				('cfg.passwd_file',         ''),             # _infile_opts - check_infile()
				('cfg.outdir',              ''),             # check_outdir()
				('cfg.cached_balances',     'False'),
				('cfg.minconf',             '1'),
				('cfg.coin',                'BTC'),
				('cfg.pager',               'False'),
				('cfg.fee_estimate_mode',   'conservative'), # _autoset_opts
			))

	def opt_good1(self):
		pf_base = 'testfile'
		pf = os.path.join(self.tmpdir, pf_base)
		self.write_to_tmpfile(pf_base, '')
		return self.check_vals(
			[
				'--print-checksum',
				'--fee-estimate-mode=E',
				'--passwd-file='+pf,
				'--outdir='+self.tmpdir,
				'--cached-balances',
				f'--hidden-incog-input-params={pf},123',
			], (
				('cfg.print_checksum',           'True'),
				('cfg.quiet',                    'True'), # set by print_checksum
				('cfg.passwd_file',              pf),
				('cfg.outdir',                   self.tmpdir),
				('cfg.cached_balances',          'True'),
				('cfg.hidden_incog_input_params', pf+',123'),
				('cfg.fee_estimate_mode',         'economical'),
			))

	def opt_good2(self):
		return self.check_vals(
			[
				'--print-checksum',
				'-qX',
				f'--outdir={self.tmpdir}',
				'-p5',
				'-m', '0',
				'--seed-len=256',
				'-L--my-label',
				'--seed-len', '128',
				'--min-temp=-30',
				'-T-10',
				'--',
				'x', 'y', '12345'
			], (
				('cfg.print_checksum',  'True'),
				('cfg.quiet',           'True'),
				('cfg.outdir',          self.tmpdir),
				('cfg.cached_balances', 'True'),
				('cfg.minconf',         '0'),
				('cfg.keep_label',      'None'),
				('cfg.seed_len',        '128'),
				('cfg.hash_preset',     '5'),
				('cfg.label',           '--my-label'),
				('cfg.min_temp',        '-30'),
				('cfg.max_temp',        '-10'),
				('arg1',                'x'),
				('arg2',                'y'),
				('arg3',                '12345'),
			))

	def opt_good3(self):
		return self.check_vals(['m'] * 256, (('arg256', 'm'),))

	def opt_good4(self):
		return self.check_vals(['e' * 4096], (('arg1', 'e' * 4096),))

	def opt_good5(self):
		return self.check_vals(['--minc=7'], (('cfg.minconf', '7'),))

	def opt_good6(self):
		if self.cfg.no_altcoin:
			return 'skip'
		return self.check_vals(['--coin=xmr'], (('cfg.coin', 'XMR'),))

	def opt_good7(self):
		if self.cfg.no_altcoin:
			return 'skip'
		return self.check_vals(['--coin', 'xmr'], (('cfg.coin', 'XMR'),))

	def opt_good8(self):
		return self.check_vals(['--pager'], (('cfg.pager', 'True'),))

	def opt_good9(self):
		return self.check_vals(['-'], (('arg1', '-'),))

	def opt_good10(self):
		return self.check_vals(['-', '-x'], (('arg1', '-'), ('arg2', '-x')))

	def opt_good11(self):
		return self.check_vals(['-q', '-', '-x'], (('arg1', '-'), ('arg2', '-x')))

	def opt_good12(self):
		return self.check_vals(['-l128'], (('cfg.seed_len', '128'),))

	def opt_good13(self):
		return self.check_vals(['-l', '128'], (('cfg.seed_len', '128'),))

	def opt_good14(self):
		return self.check_vals(['-kl128'], (('cfg.keep_label', 'True'), ('cfg.seed_len', '128')))

	def opt_good15(self):
		return self.check_vals(['-kl', '128'], (('cfg.keep_label', 'True'), ('cfg.seed_len', '128')))

	def opt_good16(self):
		return self.check_vals(['--point=x=1,y=2,z=3'], (('cfg.point', 'x=1,y=2,z=3'),))

	def opt_good17(self):
		return self.check_vals(['--point', 'x=1,y=2,z=3'], (('cfg.point', 'x=1,y=2,z=3'),))

	def opt_good18(self):
		return self.check_vals(['-xx=1,y=2,z=3'], (('cfg.point', 'x=1,y=2,z=3'),))

	def opt_good19(self):
		return self.check_vals(['-x', 'x=1,y=2,z=3'], (('cfg.point', 'x=1,y=2,z=3'),))

	def opt_good20(self):
		return self.check_vals(['--pager', '--no-pager'], (('cfg.pager', 'False'),))

	def opt_good21(self):
		return self.check_vals(['--foobleize'], (('cfg.no_foobleize', 'False'),))

	def opt_good22(self):
		return self.check_vals(['--quiet', '--no-q'], (('cfg.quiet', 'False'),))

	def opt_good23(self):
		return self.check_vals(['--foobl'], (('cfg.no_foobleize', 'False'),))

	def opt_good24(self):
		return self.check_vals(['--no-pag', '--pag'], (('cfg.pager', 'True'),))

	def opt_good25(self):
		return self.check_vals(
			['--btc-rpc-host=pi5'],
			(('cfg.btc_rpc_host', 'pi5'), ('proto.rpc_host', 'pi5')),
			need_proto=True)

	def opt_good26(self):
		return self.check_vals(
			['--btc-rpc-port=7272'],
			(('cfg.btc_rpc_port', '7272'), ('proto.rpc_port', '7272')),
			need_proto=True)

	def opt_good27(self):
		return self.check_vals(
			['--btc-ignore-daemon-version'],
			(('cfg.btc_ignore_daemon_version', 'True'), ('proto.ignore_daemon_version', 'True'),),
			need_proto = True)

	def opt_good28(self):
		return self.check_vals(
			['--coin=bch', '--bch-cashaddr=yes'],
			(('cfg.bch_cashaddr', 'True'), ('proto.cashaddr', 'True'),),
			need_proto = True)

	def opt_good29(self):
		return self.check_vals(['--etc-max-tx-fee=0.1'], (('cfg.etc_max_tx_fee', '0.1'),), need_proto=True)

	def opt_good30(self):
		if self.cfg.no_altcoin:
			return 'skip'
		return self.check_vals(
			['--coin=eth', '--eth-mainnet-chain-names=foo,bar'],
			(('cfg.eth_mainnet_chain_names', r"\['foo', 'bar'\]"), ('proto.chain_names', r"\['foo', 'bar'\]")),
			need_proto = True)

	def opt_good31(self):
		if self.cfg.no_altcoin:
			return 'skip'
		return self.check_vals(
			['--coin=xmr', '--xmr-rpc-port=28081'],
			(('cfg.xmr_rpc_port', '28081'),('proto.rpc_port', '28081'),),
			need_proto = True)

	def opt_bad_param(self):
		return self.do_run(['--pager=1'], 'no parameter', 1)

	def opt_bad_infile(self):
		pf = os.path.join(self.tmpdir, 'fubar')
		return self.do_run(['--passwd-file='+pf], 'not found', 1)

	def opt_bad_outdir(self):
		bo = self.tmpdir+'_fubar'
		return self.do_run(['--outdir='+bo], 'not found', 1)

	def opt_bad_incompatible(self):
		return self.do_run(['--label=Label', '--keep-label'], 'Conflicting options', 1)

	def opt_bad_autoset(self):
		return self.do_run(['--fee-estimate-mode=Fubar'], 'not unique substring', 1)

	def opt_invalid(self, args, expect, opts=[], need_proto=False, exit_val=1):
		t = self.spawn_prog(args, opts=opts, exit_val=exit_val, need_proto=need_proto)
		t.expect(expect)
		return t

	def opt_invalid_1(self):
		return self.opt_invalid(['--x'], 'must be at least')

	def opt_invalid_2(self):
		return self.opt_invalid(['---'], 'must be at least')

	def opt_invalid_5(self):
		return self.opt_invalid(['-l'], 'missing parameter')

	def opt_invalid_6(self):
		return self.opt_invalid(['-l', '-k'], 'missing parameter')

	def opt_invalid_7(self):
		return self.opt_invalid(['--quiet=1'], 'requires no parameter')

	def opt_invalid_8(self):
		return self.opt_invalid(['-w'], 'unrecognized option')

	def opt_invalid_9(self):
		return self.opt_invalid(['--frobnicate'], 'unrecognized option')

	def opt_invalid_10(self):
		return self.opt_invalid(['--label', '-q'], 'missing parameter')

	def opt_invalid_11(self):
		return self.opt_invalid(['-T', '-10'], 'missing parameter')

	def opt_invalid_12(self):
		return self.opt_invalid(['-q', '-10'], 'unrecognized option')

	def opt_invalid_13(self):
		return self.opt_invalid(['--mi=3'], 'ambiguous option')

	def opt_invalid_14(self):
		return self.opt_invalid(['--m=3'], 'must be at least')

	def opt_invalid_15(self):
		return self.opt_invalid(['m'] * 257, 'too many')

	def opt_invalid_16(self):
		return self.opt_invalid(['e' * 4097], 'too long')

	def opt_invalid_17(self):
		return self.opt_invalid(['--btc-rpc-host'], 'unrecognized option')

	def opt_invalid_18(self):
		return self.opt_invalid(['--btc-rpc-port'], 'unrecognized option')

	def opt_invalid_19(self):
		return self.opt_invalid(['--btc-rpc-port=foo'], "must be of type 'int'", need_proto=True)

	def opt_invalid_21(self):
		return self.opt_invalid(['--btc-foo'], 'unrecognized option')

	def opt_invalid_22(self):
		return self.opt_invalid(['--btc-rpc-host'], 'missing parameter', need_proto=True)

	def opt_invalid_23(self):
		return self.opt_invalid(['--btc-ignore-daemon-version=1'], 'requires no parameter', need_proto=True)

	def opt_invalid_24(self):
		return self.opt_invalid(['--bch-cashaddr'], 'unrecognized option')

	def opt_invalid_25(self):
		return self.opt_invalid(['--bch-cashaddr'], 'missing parameter', need_proto=True)

	def opt_invalid_26(self):
		return self.opt_invalid(['--bch-cashaddr=foo'], "must be of type 'bool'", need_proto=True)

	def opt_invalid_27(self):
		return self.opt_invalid(['--ltc-cashaddr'], 'unrecognized option', need_proto=True)

	def opt_invalid_28(self):
		return self.opt_invalid(['--xmr-max-tx-fee=0.1'], 'unrecognized option', need_proto=True)

	def opt_invalid_29(self):
		return self.opt_invalid(['--eth-max-tx-fee'], 'missing parameter', need_proto=True)

	def opt_invalid_30(self):
		return self.opt_invalid(['--eth-max-tx-fee=true'], 'must be of type', need_proto=True)

	def opt_invalid_31(self):
		return self.opt_invalid(['--bch-cashaddr=true'], 'has no attribute', opts=['--show-opts=bch_cashaddr'], need_proto=True)

	def opt_invalid_32(self):
		return self.opt_invalid(['--eth-chain-names=foo,bar'], 'unrecognized option', need_proto=True)

	def opt_invalid_33(self):
		return self.opt_invalid(['--xmr-rpc-host=solaris'], 'unrecognized option', need_proto=True)

	def opt_invalid_34(self):
		return self.opt_invalid(['--eth-rpc-user=bob'], 'unrecognized option', need_proto=True)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.ethbump: Ethereum transaction bumping tests for the cmdtest.py test suite
"""

import sys, time, asyncio, json

from mmgen.cfg import Config
from mmgen.protocol import init_proto
from mmgen.util import ymsg, suf

from ..include.common import imsg, omsg_r

from .include.common import cleanup_env, dfl_words_file, dfl_sid
from .include.runner import CmdTestRunner
from .httpd.thornode.swap import ThornodeSwapServer

from .ethdev import CmdTestEthdev, CmdTestEthdevMethods
from .regtest import CmdTestRegtest
from .swap import CmdTestSwapMethods, create_cross_methods
from .ethswap import CmdTestEthSwapMethods

burn_addr = 'beefcafe22' * 4
method_template = """
def {name}(self):
	self.spawn(log_only=True)
	return ethbump_ltc.run_test("{ltc_name}", sub=True)
"""

class CmdTestEthBumpMethods:

	@property
	def devnet_block_period(self):
		return (
			self.cfg.devnet_block_period
			or self.cfg.test_suite_devnet_block_period
			or self.dfl_devnet_block_period[self.daemon.id])

	def _txcreate(self, args, acct):
		self.get_file_with_ext('rawtx', delete_all=True)
		return self.txcreate(args, acct=acct, interactive_fee='0.9G', fee_info_data=('0.0000189', '0.9'))

	def _txsign(self, has_label=True):
		self.get_file_with_ext('sigtx', delete_all=True)
		return self.txsign(has_label=has_label)

	def _txsend(self, has_label=True):
		return self.txsend(has_label=has_label)

	def _txbump_feebump(self, *args, **kwargs):
		self.get_file_with_ext('rawtx', delete_all=True)
		return self._txbump(*args, **kwargs)

	def _txbump_new_outputs(self, *, args, fee, add_opts=[]):
		self.get_file_with_ext('rawtx', delete_all=True)
		ext = '{}.regtest.sigtx'.format('-α' if self.cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext, no_dot=True)
		return self.txbump_ui_common(
			self.spawn('mmgen-txbump', self.eth_opts + add_opts + args + [txfile]),
			fee = fee,
			fee_desc = 'or gas price',
			bad_fee = '0.9G')

	def _token_fund_user(self, *, mm_idxs):
		return self._token_transfer_ops(
			op          = 'fund_user',
			mm_idxs     = mm_idxs,
			token_addr  = 'token_addr1',
			amt         = self.token_fund_amt)

	def _token_txcreate(self, *, args, cmd='txcreate'):
		self.get_file_with_ext('sigtx', delete_all=True)
		t = self._create_token_tx(cmd=cmd, fee='1.3G', args=args, add_opts=self.eth_opts)
		t.expect('to confirm: ', 'YES\n')
		t.written_to_file('Sent transaction')
		return t

	async def _wait_for_block(self, require_seen=True):
		self.spawn(msg_only=True)
		if self.devnet_block_period:
			empty_pools_seen = 0
			tx_seen = False
			while True:
				await asyncio.sleep(1)
				t = self.spawn(
					'mmgen-cli',
					['--regtest=1', 'txpool_content'],
					env = cleanup_env(self.cfg),
					no_msg = True,
					silent = True)
				res = json.loads(t.read().strip())
				if p := res['pending']:
					imsg(f'Pool has {len(p)} transaction{suf(p)}')
					if self.tr.quiet:
						omsg_r('+')
					tx_seen = True
				else:
					imsg('Pool is empty')
					if self.tr.quiet:
						omsg_r('+')
					if tx_seen or not require_seen:
						break
					empty_pools_seen += 1
					if empty_pools_seen > 5:
						m = ('\nTransaction pool empty! Try increasing the block period with the'
							' --devnet-block-period option (current value is {})')
						ymsg(m.format(self.devnet_block_period))
						sys.exit(1)

		return 'ok'

class CmdTestEthBump(CmdTestEthBumpMethods, CmdTestEthSwapMethods, CmdTestSwapMethods, CmdTestEthdev):
	'Ethereum transaction bumping operations'

	networks = ('eth',)
	tmpdir_nums = [42]
	dfl_devnet_block_period = {'geth': 7, 'reth': 9}
	fund_amt = 100000
	token_fund_amt = 1000
	cross_group = 'ethbump_ltc'
	cross_coin = 'ltc'

	cmd_group_in = (
		('subgroup.ltc_init',           []),
		('subgroup.eth_init',           []),
		('subgroup.feebump',            ['eth_init']),
		('subgroup.new_outputs',        ['eth_init']),
		('subgroup.swap_feebump',       ['ltc_init', 'eth_init']),
		('subgroup.swap_new_outputs',   ['ltc_init', 'eth_init']),
		('subgroup.token_init',         ['eth_init']),
		('subgroup.token_feebump',      ['token_init']),
		('subgroup.token_new_outputs',  ['token_init']),
		('subgroup.token_init_swap',    ['token_init']),
		# ('subgroup.token_feebump_swap',  ['token_init_swap']), # TBD
		('subgroup.token_new_outputs_swap',  ['token_init_swap']),
		('ltc_stop',                    ''),
		('swap_server_stop',            'stopping the Thornode server'),
		('stop',                        'stopping daemon'),
	)
	cmd_subgroups = {
		'eth_init': (
			'initializing ETH tracking wallet',
			('setup',               'dev mode transaction bumping tests for Ethereum (start daemon)'),
			('addrgen',             'generating addresses'),
			('addrimport',          'importing addresses'),
			('addrimport_devaddr',  'importing the dev address'),
			('addrimport_reth_devaddr','importing the reth dev address'),
			('fund_devaddr',        'funding the dev address'),
			('wait_reth1',          'waiting for block'),
			('del_reth_devaddr',    'deleting the reth dev address'),
			('fund_mmgen_addr1',    'funding user address :1)'),
			('fund_mmgen_addr2',    'funding user address :11)'),
			('fund_mmgen_addr3',    'funding user address :21)'),
			('wait1',               'waiting for block'),
		),
		'ltc_init': (
			'initializing LTC tracking wallet',
			('ltc_setup',          ''),
			('ltc_walletconv_bob', ''),
			('ltc_addrgen_bob',    ''),
			('ltc_addrimport_bob', ''),
		),
		'feebump': (
			'creating, signing, sending, bumping and resending a transaction (fee-bump only)',
			('txcreate1',   'creating a transaction (send to burn address)'),
			('txsign1',     'signing the transaction'),
			('txsend1',     'sending the transaction'),
			('txbump1',     'creating a replacement transaction (fee-bump)'),
			('txbump1sign', 'signing the replacement transaction'),
			('txbump1send', 'sending the replacement transaction'),
			('wait2',       'waiting for block'),
			('bal1',        'checking the balance'),
		),
		'new_outputs': (
			'creating, signing, sending, bumping and resending a transaction (new outputs)',
			('txcreate2',   'creating a transaction (send to burn address)'),
			('txsign2',     'signing the transaction'),
			('txsend2',     'sending the transaction'),
			('txbump2',     'creating a replacement transaction (new outputs)'),
			('txbump2sign', 'signing the replacement transaction'),
			('txbump2send', 'sending the replacement transaction'),
			('wait3',       'waiting for block'),
			('bal2',        'checking the balance'),
		),
		'swap_feebump': (
			'creating, signing, sending, bumping and resending a swap transaction (fee-bump only)',
			('swaptxcreate1',   'creating a swap transaction (from address :11)'),
			('swaptxsign1',     'signing the transaction'),
			('swaptxsend1',     'sending the transaction'),
			('swaptxbump1',     'creating a replacement swap transaction (fee-bump)'),
			('swaptxbump1sign', 'signing the replacement transaction'),
			('swaptxbump1send', 'sending the replacement transaction'),
			('wait4',           'waiting for block'),
			('bal3',            'checking the balance'),
		),
		'swap_new_outputs': (
			'creating, signing, sending, bumping and resending a swap transaction (new output)',
			('swaptxcreate2',   'creating a swap transaction (from address :21)'),
			('swaptxsign2',     'signing the transaction'),
			('swaptxsend2',     'sending the transaction'),
			('swaptxbump2',     'creating a replacement swap transaction (new output)'),
			('swaptxbump2sign', 'signing the replacement transaction'),
			('swaptxbump2send', 'sending the replacement transaction'),
			('wait5',           'waiting for block'),
			('bal4',            'checking the balance'),
		),
		'token_init': (
			'initializing token wallets',
			('token_compile1',   'compiling ERC20 token #1'),
			('token_deploy_a',   'deploying ERC20 token MM1 (SafeMath)'),
			('token_deploy_b',   'deploying ERC20 token MM1 (Owned)'),
			('token_deploy_c',   'deploying ERC20 token MM1 (Token)'),
			('token_fund_user1', 'transferring token funds from dev to user (addr #1)'),
			('token_addrgen',    'generating token addresses'),
			('token_addrimport', 'importing token addresses using token address (MM1)'),
			('token_bal1',       'the token balance'),
		),
		'token_feebump': (
			'creating, signing, sending, bumping and resending a token transaction (fee-bump only)',
			('token_txdo1',   'creating, signing and sending a token transaction'),
			('token_txbump1', 'bumping the token transaction (fee-bump)'),
			('wait6',         'waiting for block'),
			('token_bal2',    'the token balance'),
		),
		'token_new_outputs': (
			'creating, signing, sending, bumping and resending a token transaction (new outputs)',
			('token_txdo2',       'creating, signing and sending a token transaction'),
			('token_txbump2',     'creating a replacement token transaction (new outputs)'),
			('token_txbump2sign', 'signing the replacement transaction'),
			('token_txbump2send', 'sending the replacement transaction'),
			('wait7',             'waiting for block'),
			('token_bal3',        'the token balance'),
		),
		'token_init_swap': (
			'initializing token swap configuration',
			('token_compile_router',  'compiling THORChain router contract'),
			('token_deploy_router',   'deploying THORChain router contract'),
		),
		'token_feebump_swap': (
			'creating, signing, sending, bumping and resending a token swap transaction (feebump)',
			('token_fund_user11',     'transferring token funds from dev to user (addr #11)'),
			('token_addrimport_inbound', 'importing THORNode inbound token address'),
			('token_swaptxdo1',       'creating, signing and sending a token transaction (feebump)'),
			('token_swaptxbump1',     'bumping the token transaction (fee-bump)'),
			('token_swaptxbump1sign', 'signing the replacement transaction'),
			('token_swaptxbump1send', 'sending the replacement transaction'),
			('wait8',                 'waiting for block'),
			('token_bal5',            'the token balance'),
		),
		'token_new_outputs_swap': (
			'creating, signing, sending, bumping and resending a token swap transaction (new outputs)',
			('token_swaptxdo2',       'creating, signing and sending a token swap transaction (new outputs)'),
			('token_swaptxbump2',     'creating a replacement token transaction'),
			('token_swaptxbump2sign', 'signing the replacement transaction'),
			('token_swaptxbump2send', 'sending the replacement transaction'),
			('wait9',                 'waiting for block'),
			('token_bal6',            'the token balance'),
		),
	}

	exec(create_cross_methods(cross_coin, cross_group, cmd_group_in, cmd_subgroups))

	def __init__(self, cfg, trunner, cfgs, spawn):

		CmdTestEthdev.__init__(self, cfg, trunner, cfgs, spawn)

		if not trunner:
			return

		self.daemon.usr_coind_args = {
			'reth': [f'--dev.block-time={self.devnet_block_period}s'],
			'geth': [f'--dev.period={self.devnet_block_period}']
		}[self.daemon.id]

		imsg(f'devnet block period: {self.devnet_block_period}')

		globals()[self.cross_group] = self.create_cross_runner(trunner)

		self.swap_server = ThornodeSwapServer()
		self.swap_server.start()

	def txcreate1(self):
		return self._txcreate(args=[f'{burn_addr},987'], acct='1')

	def txbump1(self):
		return self._txbump_feebump(fee='1.3G', ext='{}.regtest.sigtx')

	def txcreate2(self):
		return self._txcreate(args=[f'{burn_addr},789'], acct='1')

	def txbump2(self):
		return self._txbump_new_outputs(args=[f'{dfl_sid}:E:2,777'], fee='1.3G')

	def swaptxcreate1(self):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(['ETH', '12.34567', 'LTC', f'{dfl_sid}:B:3']),
			inputs = 4)

	def swaptxsign1(self):
		return self._swaptxsign()

	def swaptxsend1(self):
		return self._swaptxsend()

	def swaptxbump1(self):
		return self._swaptxbump('41.1G')

	def swaptxbump2(self):
		return self._swaptxbump('1.9G', output_args=[f'{dfl_sid}:E:12,4444.3333'])

	def bal1(self):
		return self._bal_check(pat=rf'{dfl_sid}:E:1\s+99012\.9999727\s')

	def bal2(self):
		return self._bal_check(pat=rf'{dfl_sid}:E:2\s+777\s')

	def bal3(self):
		dec = {'geth': '653431389777251448', 'reth': '65337812418775812'}[self.daemon.id]
		return self._bal_check(pat=rf'{dfl_sid}:E:11\s+99987\.{dec}\s')

	def bal4(self):
		return self._bal_check(pat=rf'{dfl_sid}:E:12\s+4444\.3333\s')

	def token_fund_user1(self):
		return self._token_fund_user(mm_idxs=[1])

	def token_fund_user11(self):
		return self._token_fund_user(mm_idxs=[11])

	def token_txdo1(self):
		return self._token_txcreate(cmd='txdo', args=[f'{dfl_sid}:E:2,1.23456', dfl_words_file])

	def token_txbump1(self):
		t = self._txbump_feebump(
			fee = '60G',
			ext = '{}.regtest.sigtx',
			add_opts = ['--token=MM1'],
			add_args = [dfl_words_file])
		t.expect('to confirm: ', 'YES\n')
		t.written_to_file('Sent transaction')
		return t

	def token_bal2(self):
		return self._token_bal_check(pat=rf'{dfl_sid}:E:2\s+1\.23456')

	def token_txdo2(self):
		return self._token_txcreate(cmd='txdo', args=[f'{dfl_sid}:E:3,5.4321', dfl_words_file])

	def token_txbump2(self):
		return self._txbump_new_outputs(
			args = [f'{dfl_sid}:E:4,6.54321'],
			fee = '1.6G',
			add_opts = ['--token=mm1', '--gas=75000'])

	def token_txbump2sign(self):
		return self._txsign(has_label=False)

	def token_txbump2send(self):
		return self._txsend(has_label=False)

	def token_bal3(self):
		return self._token_bal_check(pat=rf'{dfl_sid}:E:4\s+6\.54321')

	def wait_reth1(self):
		return self._wait_for_block() if self.daemon.id == 'reth' else 'silent'

	def token_swaptxdo1(self):
		self.get_file_with_ext('sigtx', delete_all=True)
		t = self._swaptxcreate(
				['ETH.MM1', '0.321', 'ETH', dfl_words_file],
				action = 'txdo')
		t.expect('(Y/n): ', '\n')
		return self._swaptxcreate_ui_common(
			t,
			sign_and_send = True,
			need_passphrase = False,
			file_desc = 'Sent transaction',
			inputs = 11)

	def token_swaptxbump1(self):
		time.sleep(0.2)
		self.get_file_with_ext('rawtx', delete_all=True)
		txfile = self.get_file_with_ext('sigtx', no_dot=True)
		t = self.spawn(
			'mmgen-txbump',
			self.eth_opts
			+ ['--gas=50000'] # , '--router-gas=600000']
			+ ['--yes', txfile])
		t.expect('to continue: ', '\n')     # exit swap quote view
		t.expect('or gas price: ', '8G\n')  # enter fee
		t.expect(r'Gas limit:.*\D650000\D', regex=True)
		t.written_to_file('Fee-bumped transaction')
		return t

	def token_swaptxdo2(self):
		self.get_file_with_ext('sigtx', delete_all=True)
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(
				['ETH.MM1', '0.321', 'ETH', f'{dfl_sid}:E:21', dfl_words_file],
				action = 'txdo'),
			sign_and_send = True,
			need_passphrase = False,
			file_desc = 'Sent transaction',
			inputs = 1)

	def token_swaptxbump2(self):
		return self._txbump_new_outputs(
			args = [f'{dfl_sid}:E:8,0.54321'],
			fee = '1.4G',
			add_opts = ['--gas=67888', '--fee=3G'])

	def token_bal5(self):
		return self._token_bal_check(pat=r'feedbeefcafe\s+non-MMGen\s+0\.321\s')

	def token_bal6(self):
		return self._token_bal_check(pat=rf'{dfl_sid}:E:8\s+0\.54321')

	wait1 = wait2 = wait3 = wait4 = wait5 = wait6 = wait7 = wait8 = wait9 = CmdTestEthBumpMethods._wait_for_block

	txsign1 = txsign2 = txbump1sign = txbump2sign = CmdTestEthBumpMethods._txsign
	txsend1 = txsend2 = txbump1send = txbump2send = CmdTestEthBumpMethods._txsend

	swaptxcreate2 = swaptxcreate1
	swaptxsign2 = swaptxsign1
	swaptxsend2 = swaptxsend1

	token_swaptxbump1sign = token_swaptxbump2sign = swaptxbump1sign = swaptxbump2sign = token_txbump2sign
	token_swaptxbump1send = token_swaptxbump2send = swaptxbump1send = swaptxbump2send = token_txbump2send

	def swap_server_stop(self):
		return self._thornode_server_stop()

class CmdTestEthBumpLTC(CmdTestSwapMethods, CmdTestRegtest):
	'Ethereum transaction bumping operations - LTC wallet'
	network = ('ltc',)
	tmpdir_nums = [43]
	is_helper = True
	cmd_group_in = CmdTestRegtest.cmd_group_in + (
		('setup',           'LTC regtest setup'),
		('walletconv_bob',  'LTC wallet generation'),
		('addrgen_bob',     'LTC address generation'),
		('addrimport_bob',  'importing LTC addresses'),
		('stop',            'stopping the Litecoin daemon'),
	)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
mmgen-msg: Message signing operations for the MMGen suite
"""

import sys
from .cfg import Config
from .base_obj import AsyncInit
from .util import msg, suf, async_run, die
from .msg import (
	NewMsg,
	UnsignedMsg,
	SignedMsg,
	SignedOnlineMsg,
	ExportedMsgSigs,
)

class MsgOps:
	ops = ('create', 'sign', 'verify')

	class create:

		def __init__(self, msg, addr_specs):
			NewMsg(
				cfg       = cfg,
				coin      = cfg._proto.coin,
				network   = cfg._proto.network,
				message   = msg,
				addrlists = addr_specs,
				msghash_type = cfg.msghash_type
			).write_to_file(ask_overwrite=False)

	class sign(metaclass=AsyncInit):

		async def __init__(self, msgfile, wallet_files):

			m = UnsignedMsg(cfg, infile=msgfile)

			if not wallet_files:
				from .filename import find_file_in_dir
				from .wallet import get_wallet_cls
				wallet_files = [find_file_in_dir(get_wallet_cls('mmgen'), cfg.data_dir)]

			await m.sign(wallet_files)

			m = SignedMsg(cfg, data=m.__dict__)

			m.write_to_file(ask_overwrite=False)

			if m.data.get('failed_sids'):
				sys.exit(1)

	class verify(sign):

		async def __init__(self, msgfile, *, addr=None):
			try:
				m = SignedOnlineMsg(cfg, infile=msgfile)
			except:
				m = ExportedMsgSigs(cfg, infile=msgfile)

			nSigs = await m.verify(addr=addr)

			summary = f'{nSigs} signature{suf(nSigs)} verified'

			if cfg.quiet:
				msg(summary)
			else:
				cfg._util.stdout_or_pager(m.format(addr) + '\n\n' + summary + '\n')

			if m.data.get('failed_sids'):
				sys.exit(1)

	class export(sign):

		async def __init__(self, msgfile, *, addr=None):

			from .fileutil import write_data_to_file
			write_data_to_file(
				cfg     = cfg,
				outfile = 'signatures.json',
				data    = SignedOnlineMsg(cfg, infile=msgfile).get_json_for_export(addr=addr),
				desc    = 'signature data')

opts_data = {
	'text': {
		'desc': 'Perform message signing operations for MMGen addresses',
		'usage2': [
			'[opts] create MESSAGE_TEXT ADDRESS_SPEC [...]',
			'[opts] sign   MESSAGE_FILE [WALLET_FILE ...]',
			'[opts] verify MESSAGE_FILE [MMGen ID]',
			'[opts] verify <exported JSON dump file> [address]',
			'[opts] export MESSAGE_FILE [MMGen ID]',
		],
		'options': """
-h, --help           Print this help message
--, --longhelp       Print help message for long (global) options
-d, --outdir=d       Output file to directory 'd' instead of working dir
-t, --msghash-type=T Specify the message hash type.  Supported values:
                     'eth_sign' (ETH default), 'raw' (non-ETH default)
-q, --quiet          Produce quieter output
""",
	'notes': """

                             SUPPORTED OPERATIONS

create - create a raw MMGen message file with specified message text for
         signing for addresses specified by ADDRESS_SPEC (see ADDRESS
         SPECIFIER below)
sign   - perform signing operation on an unsigned MMGen message file
verify - verify and display the contents of a signed MMGen message file
export - dump signed MMGen message file to ‘signatures.json’, including only
         data relevant for a third-party verifier


                              ADDRESS SPECIFIER

The `create` operation takes one or more ADDRESS_SPEC arguments with the
following format:

    SEED_ID:ADDRTYPE_CODE:ADDR_IDX_SPEC

where ADDRTYPE_CODE is a one-letter address type code from the list below, and
ADDR_IDX_SPEC is a comma-separated list of address indexes or hyphen-separated
address index ranges.

{n_at}


                                    NOTES

Message signing operations are supported for Bitcoin, Ethereum and code forks
thereof.

By default, Ethereum messages are prefixed before hashing in conformity with
the standard defined by the Geth ‘eth_sign’ JSON-RPC call.  This behavior may
be overridden with the --msghash-type option.

Messages signed for Segwit-P2SH addresses cannot be verified directly using
the Bitcoin Core `verifymessage` RPC call, since such addresses are not hashes
of public keys.  As a workaround for this limitation, this utility creates for
each Segwit-P2SH address a non-Segwit address with the same public key to be
used for verification purposes.  This non-Segwit verifying address should then
be passed on to the verifying party together with the signature. The verifying
party may then use a tool of their choice (e.g. `mmgen-tool addr2pubhash`) to
assure themselves that the verifying address and Segwit address share the same
public key.

Unfortunately, the aforementioned limitation applies to Segwit-P2PKH (Bech32)
addresses as well, despite the fact that Bech32 addresses are hashes of public
keys (we consider this an implementation shortcoming of `verifymessage`).
Therefore, the above procedure must be followed to verify messages for Bech32
addresses too.  `mmgen-tool addr2pubhash` or `bitcoin-cli validateaddress`
may then be used to demonstrate that the two addresses share the same public
key.


                                   EXAMPLES

Create a raw message file for the specified message and specified addresses,
where DEADBEEF is the Seed ID of the user’s default wallet and BEEFCAFE one
of its subwallets:
$ mmgen-msg create '16/3/2022 Earthquake strikes Fukushima coast' DEADBEEF:B:1-3,10,98 BEEFCAFE:S:3,9

Sign the raw message file created by the previous step:
$ mmgen-msg sign <raw message file>

Sign the raw message file using an explicitly supplied wallet:
$ mmgen-msg sign <raw message file> DEADBEEF.bip39

Verify and display all signatures in the signed message file:
$ mmgen-msg verify <signed message file>

Verify and display a single signature in the signed message file:
$ mmgen-msg verify <signed message file> DEADBEEF:B:98

Export data relevant for a third-party verifier to ‘signatures.json’:
$ mmgen-msg export <signed message file>

Same as above, but export only one signature:
$ mmgen-msg export <signed message file> DEADBEEF:B:98

Verify and display the exported JSON signature data:
$ mmgen-msg verify signatures.json
"""
	},
	'code': {
		'notes': lambda help_notes, s: s.format(
			n_at = help_notes('address_types'),
		)
	}
}

cfg = Config(opts_data=opts_data, need_amt=False)

cmd_args = cfg._args

if len(cmd_args) < 2:
	cfg._usage()

op = cmd_args.pop(0)

if cfg.msghash_type and op != 'create':
	die(1, '--msghash-type option may only be used with the "create" command')

async def main():
	if op == 'create':
		if len(cmd_args) < 2:
			cfg._usage()
		MsgOps.create(cmd_args[0], ' '.join(cmd_args[1:]))
	elif op == 'sign':
		if len(cmd_args) < 1:
			cfg._usage()
		await MsgOps.sign(cmd_args[0], cmd_args[1:])
	elif op in ('verify', 'export'):
		if len(cmd_args) not in (1, 2):
			cfg._usage()
		await getattr(MsgOps, op)(cmd_args[0], addr=cmd_args[1] if len(cmd_args) == 2 else None)
	else:
		die(1, f'{op!r}: unrecognized operation')

async_run(main())
